#!python3
# -*- coding: utf-8 -*-
"""

"""
from typing import List, Union, Tuple, Dict
from collections import OrderedDict
import copy
import re
from warnings import warn
import argparse
import io
import hashlib

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.auxiliary.ckan_auxiliary import json_encode_params
from ckanapi_harvesters.auxiliary.ckan_configuration import default_ckan_has_postgis, default_ckan_target_epsg
from ckanapi_harvesters.auxiliary.proxy_config import ProxyConfig
from ckanapi_harvesters.auxiliary.ckan_model import CkanPackageInfo, CkanResourceInfo, CkanViewInfo, CkanField
from ckanapi_harvesters.auxiliary.ckan_model import CkanState
from ckanapi_harvesters.auxiliary.ckan_auxiliary import assert_or_raise, ckan_package_name_re, datastore_id_col
from ckanapi_harvesters.auxiliary.ckan_auxiliary import dict_recursive_update
from ckanapi_harvesters.auxiliary.ckan_auxiliary import upload_prepare_requests_files_arg, RequestType
from ckanapi_harvesters.auxiliary.ckan_action import CkanNotFoundError
from ckanapi_harvesters.auxiliary.ckan_errors import (ReadOnlyError, AdminFeatureLockedError, NoDefaultView,
                                                      InvalidParameterError, CkanMandatoryArgumentError,
                                                      IntegrityError)
from ckanapi_harvesters.policies.data_format_policy import CkanPackageDataFormatPolicy
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_abc import CkanDataCleanerABC
from ckanapi_harvesters.ckan_api.ckan_api_1_map import use_ckan_owner_org_as_default

from ckanapi_harvesters.auxiliary.ckan_map import CkanMap
from ckanapi_harvesters.auxiliary.ckan_api_key import CkanApiKey
from ckanapi_harvesters.ckan_api.ckan_api_4_readwrite import CkanApiReadWriteParams
from ckanapi_harvesters.ckan_api.ckan_api_4_readwrite import CkanApiReadWrite


default_alias_package_resource_sep:str = "."
ckan_table_name_max_len:int = 63  # this comes from a PostgreSQL length limitation and applies to DataStore aliases
alias_name_max_len:Union[int,None] = ckan_table_name_max_len
# if name exceeds max length, a hash of the full name is used so this hash should be unchanged if the resource is re-created:
default_alias_hash_replace:bool = True  # True: replace with full hash / False: only the exceeding characters are replaced, with the following parameters:
default_alias_hash_len:int = 6
default_alias_hash_sep:str = ":"

table_name_subs_re = '[^\w-]|^(?=\d)'

def clean_table_name(variable_name: str) -> str:
    """
    Replace unwanted characters and spaces to generate a table name similar to a table name
    """
    return re.sub(table_name_subs_re,'_', variable_name)


class CkanApiManageParams(CkanApiReadWriteParams):
    default_enable_admin: bool = False  # False: disable advanced admin operations by default such as resource/package deletion
    default_alias_enforce: bool = True  # if True, always add the default alias when calling datastore_create

    def __init__(self, *, proxies:Union[str,dict,ProxyConfig]=None,
                 ckan_headers:dict=None, http_headers:dict=None):
        super().__init__(proxies=proxies, ckan_headers=ckan_headers, http_headers=http_headers)
        self.enable_admin: bool = self.default_enable_admin

    def copy(self, new_identifier:str=None, *, dest=None):
        if dest is None:
            dest = CkanApiManageParams()
        super().copy(dest=dest)
        dest.enable_admin = self.enable_admin
        return dest

    def _setup_cli_ckan_parser__params(self, parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        # overload adding support to trigger admin mode
        parser = super()._setup_cli_ckan_parser__params(parser=parser)
        parser.add_argument("--admin", action="store_true",
                            help="Option to enable admin mode")
        return parser

    def _cli_ckan_args_apply(self, args: argparse.Namespace, *, base_dir:str=None,
                             error_not_found:bool=True, default_proxies:dict=None, proxy_headers:dict=None,
                             proxies:dict=None, headers:dict=None) -> None:
        # overload adding support to trigger admin mode
        super()._cli_ckan_args_apply(args=args, base_dir=base_dir, error_not_found=error_not_found,
                                     default_proxies=default_proxies, proxy_headers=proxy_headers)
        if args.admin:
            self.enable_admin = args.admin


class CkanApiExtendedParams(CkanApiManageParams):
    def __init__(self, *, proxies:Union[str,dict,ProxyConfig]=None,
                 ckan_headers:dict=None, http_headers:dict=None):
        super().__init__(proxies=proxies, ckan_headers=ckan_headers, http_headers=http_headers)
        self.ckan_has_postgis: bool = default_ckan_has_postgis
        self.ckan_default_target_epsg: Union[int,None] = default_ckan_target_epsg

    def copy(self, new_identifier:str=None, *, dest=None):
        if dest is None:
            dest = CkanApiExtendedParams()
        super().copy(dest=dest)
        dest.ckan_has_postgis = self.ckan_has_postgis
        dest.ckan_default_target_epsg = self.ckan_default_target_epsg
        return dest

    def _setup_cli_ckan_parser__params(self, parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        # overload adding support to change extended parameters
        parser = super()._setup_cli_ckan_parser__params(parser=parser)
        parser.add_argument("--ckan-postgis", action="store_true",
                            help="Option to notify that CKAN is compatible with PostGIS")  # default=default_ckan_has_postgis
        parser.add_argument("--ckan-epsg", type=int,
                            help="Default EPSG for CKAN", default=default_ckan_target_epsg)
        return parser

    def _cli_ckan_args_apply(self, args: argparse.Namespace, *, base_dir:str=None,
                             error_not_found:bool=True, default_proxies:dict=None, proxy_headers:dict=None,
                             proxies:dict=None, headers:dict=None) -> None:
        # overload adding support to trigger admin mode
        super()._cli_ckan_args_apply(args=args, base_dir=base_dir, error_not_found=error_not_found,
                                     default_proxies=default_proxies, proxy_headers=proxy_headers)
        if args.ckan_postgis:
            self.ckan_has_postgis = args.ckan_postgis
        if args.ckan_epsg:
            self.ckan_default_target_epsg = args.ckan_epsg


class CkanApiManage(CkanApiReadWrite):
    """
    CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.
    This class implements more advanced requests to manage packages, resources and DataStores on the CKAN server.
    """

    def __init__(self, url:str=None, *, proxies:Union[str,dict,ProxyConfig]=None,
                 apikey:Union[str,CkanApiKey]=None, apikey_file:str=None,
                 owner_org:str=None, params:CkanApiExtendedParams=None,
                 map:CkanMap=None, policy: CkanPackageDataFormatPolicy = None, policy_file:str=None,
                 data_cleaner_upload:CkanDataCleanerABC=None,
                 identifier=None):
        """
        CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.

        :param url: url of the CKAN server
        :param proxies: proxies to use for requests
        :param apikey: way to provide the API key directly (optional)
        :param apikey_file: path to a file containing a valid API key in the first line of text (optional)
        :param policy: data format policy to use with policy_check function
        :param policy_file: path to a JSON file containing the data format policy to use with policy_check function
        :param owner_org: name of the organization to limit package_search (optional)
        :param params: other connection/behavior parameters
        :param map: map of known resources
        :param policy: data format policy to be used with the policy_check function.
        :param policy_file: path to a JSON file containing the data format policy to load.
        :param data_cleaner_upload: data cleaner object to use before uploading to a CKAN DataStore.
        :param identifier: identifier of the ckan client
        """
        super().__init__(url=url, proxies=proxies, apikey=apikey, apikey_file=apikey_file,
                         owner_org=owner_org, map=map, policy=policy, policy_file=policy_file,
                         data_cleaner_upload=data_cleaner_upload, identifier=identifier)
        if params is None:
            params = CkanApiExtendedParams()
        if proxies is not None:
            params.proxies = proxies
        self.params: CkanApiExtendedParams = params

    def copy(self, new_identifier: str = None, *, dest=None):
        if dest is None:
            dest = CkanApiManage()
        super().copy(new_identifier=new_identifier, dest=dest)
        return dest

    def full_unlock(self, unlock:bool=True,
                    *, no_ca:bool=None, external_url_resource_download:bool=None) -> None:
        """
        Function to unlock full capabilities of the CKAN API

        :param unlock:
        :return:
        """
        super().full_unlock(unlock, no_ca=no_ca, external_url_resource_download=external_url_resource_download)
        self.params.enable_admin = unlock

    def _setup_cli_ckan_parser(self, parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        # overload adding support to load a policy from a file
        parser = super()._setup_cli_ckan_parser(parser=parser)
        return parser

    def _cli_ckan_args_apply(self, args: argparse.Namespace, *, base_dir:str=None,
                             error_not_found:bool=True, default_proxies:dict=None, proxy_headers:dict=None,
                             proxies:dict=None, headers:dict=None) -> None:
        # overload adding support to load a policy from a file
        super()._cli_ckan_args_apply(args=args, base_dir=base_dir, error_not_found=error_not_found,
                                     default_proxies=default_proxies, proxy_headers=proxy_headers)

    ## Field modification ------------------
    @staticmethod
    def datastore_field_dict(fields:Union[List[Union[CkanField,dict]], OrderedDict[str,Union[CkanField,dict]]]=None,
                             fields_merge:Union[List[Union[CkanField,dict]], OrderedDict[str,Union[CkanField,dict]]]=None,
                             fields_update:Union[List[Union[CkanField,dict]], OrderedDict[str,Union[CkanField,dict]]]=None, *,
                             fields_type_override:Dict[str,str]=None, fields_description:Dict[str,str]=None,
                             fields_label:Dict[str,str]=None, return_list:bool=False) \
            -> Union[Dict[str, CkanField], List[dict]]:
        """
        Initialization of the `fields` parameter for datastore_create.
        Only parts used by this package are present.
        To complete the field's dictionnaries, refer to datastore_field_patch_dict.

        :param fields: first source of field information, usually the fields from the DataStore
        :param fields_merge: second source. Values from this dictionary will overwrite fields
        :param fields_update: third source. Values from this dictionary will be prioritary over all values.
        :param fields_type_override:
        :param fields_description:
        :param fields_label:
        :param return_list:
        :return: dict if return_list is False, list if return_list is True.
        You can easily transform the dict to a list with the following code:
        ```python
        fields = list(fields_update.values())
        ```
        """
        fields_updated: OrderedDict[str, CkanField]
        if fields is None:
            fields_updated = OrderedDict()
        elif isinstance(fields, list):
            fields_updated = OrderedDict()
            for field_info in fields:
                if isinstance(field_info, CkanField):
                    fields_updated[field_info.name] = field_info
                elif isinstance(field_info, dict):
                    fields_updated[field_info["id"]] = CkanField.from_ckan_dict(field_info)
                else:
                    raise TypeError(f"Field type {type(field_info)} not supported.")
        elif isinstance(fields, dict):
            fields_updated = OrderedDict()
            for field_name, field_info in fields.items():
                if isinstance(field_info, CkanField):
                    fields_updated[field_info.name] = field_info
                    assert_or_raise(field_name == field_info.name, IntegrityError(f"Field {field_name} does not match its id ({field_info.name})"))
                elif isinstance(field_info, dict):
                    fields_updated[field_info["id"]] = CkanField.from_ckan_dict(field_info)
                    assert_or_raise(field_name == field_info["id"], IntegrityError(f"Field {field_name} does not match its id ({field_info['id']})"))
                else:
                    raise TypeError(f"Field type {type(field_info)} not supported for {field_name}.")
        else:
            raise TypeError(f"Field type {type(fields)} not supported.")
        if fields_merge is None:
            pass
        elif isinstance(fields_merge, list):
            for field_info in fields_merge:
                if isinstance(field_info, CkanField):
                    fields_updated[field_info.name] = field_info
                elif isinstance(field_info, dict):
                    fields_updated[field_info["id"]] = CkanField.from_ckan_dict(field_info)
                else:
                    raise TypeError(f"Field type {type(field_info)} not supported.")
        elif isinstance(fields_merge, dict):
            for field_name, field_info in fields_merge.items():
                if isinstance(field_info, CkanField):
                    fields_updated[field_info.name] = field_info
                    assert_or_raise(field_name == field_info.name, IntegrityError(f"Field {field_name} does not match its id ({field_info.name})"))
                elif isinstance(field_info, dict):
                    fields_updated[field_info["id"]] = CkanField.from_ckan_dict(field_info)
                    assert_or_raise(field_name == field_info["id"], IntegrityError(f"Field {field_name} does not match its id ({field_info['id']})"))
                else:
                    raise TypeError(f"Field type {type(field_info)} not supported for {field_name}.")
        else:
            raise TypeError(f"Field type {type(fields_merge)} not supported.")
        if fields_update is None:
            pass
        elif isinstance(fields_update, list):
            for field_info in fields_update:
                if isinstance(field_info, CkanField):
                    fields_updated[field_info.name] = field_info
                elif isinstance(field_info, dict):
                    fields_updated[field_info["id"]] = CkanField.from_ckan_dict(field_info)
                else:
                    raise TypeError(f"Field type {type(field_info)} not supported.")
        elif isinstance(fields_update, dict):
            for field_name, field_info in fields_update.items():
                if isinstance(field_info, CkanField):
                    fields_updated[field_info.name] = field_info
                    assert_or_raise(field_name == field_info.name, IntegrityError(f"Field {field_name} does not match its id ({field_info.name})"))
                elif isinstance(field_info, dict):
                    fields_updated[field_info["id"]] = CkanField.from_ckan_dict(field_info)
                    assert_or_raise(field_name == field_info["id"], IntegrityError(f"Field {field_name} does not match its id ({field_info['id']})"))
                else:
                    raise TypeError(f"Field type {type(field_info)} not supported for {field_name}.")
        else:
            raise TypeError(f"Field type {type(fields_update)} not supported.")
        if fields_type_override is not None:
            for field_name, field_type in fields_type_override.items():
                if field_name in fields_updated.keys():
                    fields_updated[field_name].type_override = field_type
                else:
                    fields_updated[field_name] = CkanField(field_name, field_type)
            # fields_updated = dict_recursive_update(fields_updated, {field_id: {"type": str(value), "info": {"type_override": str(value)}, "schema": {"native_type": str(value)}} for field_id, value in fields_type_override.items()})
        if fields_description is not None:
            for field_name, description in fields_description.items():
                if field_name not in fields_updated.keys():
                    fields_updated[field_name] = CkanField(field_name, None)
                fields_updated[field_name].notes = description
            # fields_updated = dict_recursive_update(fields_updated, {field_id: {"info": {"notes": value}} for field_id, value in fields_description.items()})
        if fields_label is not None:
            for field_name, label in fields_description.items():
                if field_name not in fields_updated.keys():
                    fields_updated[field_name] = CkanField(field_name, None)
                fields_updated[field_name].label = label
            # fields_updated = dict_recursive_update(fields_updated, {field_id: {"info": {"label": value}} for field_id, value in fields_label.items()})
        for field_id, field_dict in fields_updated.items():
            field_dict.name = field_id
        if return_list:
            return [field_info.to_ckan_dict() for field_info in fields_updated.values()]
        else:
            return fields_updated

    def datastore_field_patch_dict(self, fields_merge:Union[List[Union[CkanField,dict]], OrderedDict[str,Union[CkanField,dict]]]=None,
                                   fields_update:Union[List[Union[CkanField,dict]], OrderedDict[str,Union[CkanField,dict]]]=None, *,
                                   fields_type_override:Dict[str,str]=None, fields_description:Dict[str,str]=None,
                                   fields_label:Dict[str,str]=None, return_list:bool=False,
                                   datastore_merge:bool=True, resource_id:str=None, error_not_found:bool=True) \
            -> Tuple[Union[bool,None], Union[Dict[str, CkanField], List[dict]]]:
        """
        Calls datastore_field_dict and merges attributes with those found in datastore_info if datastore_merge=True.

        :param fields_update:
        :param fields_type_override:
        :param fields_description:
        :param fields_label:
        :param return_list:
        :param datastore_merge:
        :param resource_id: required if datastore_merge=True
        :return:
        """
        fields_update: Dict[str, CkanField] = CkanApiManage.datastore_field_dict(fields=None, fields_merge=fields_merge, fields_update=fields_update,
                                               fields_type_override=fields_type_override, fields_description=fields_description,
                                               fields_label=fields_label, return_list=False)
        if datastore_merge:
            if error_not_found:
                assert(resource_id is not None)
            datastore_info = self.get_datastore_info_or_request_of_id(resource_id, error_not_found=error_not_found)
            if datastore_info is not None:
                fields_base = copy.deepcopy(datastore_info.fields_dict)
                if len(fields_base) == 0:
                    msg = f"No fields found for {resource_id}"
                    warn(msg)
                fields_new = copy.deepcopy(fields_base)
                update_needed = False
                for field_name, field_info in fields_update.items():
                    if field_name not in fields_base.keys():
                        fields_new[field_name] = field_info
                        update_needed = True
                    else:
                        fields_new[field_name] = fields_base[field_name].merge(field_info)
                        update_needed |= not fields_new[field_name] == fields_base[field_name]
            else:
                fields_new = fields_update
                update_needed = False
            if return_list:
                return update_needed, [field_info.to_ckan_dict() for field_info in fields_new.values()]
            else:
                return update_needed, fields_new
        else:
            if return_list:
                return None, [field_info.to_ckan_dict() for field_info in fields_update.values()]
            else:
                return None, fields_update

    def datastore_field_patch(self, resource_id:str, fields_merge:Union[List[Union[CkanField,dict]], OrderedDict[str,Union[CkanField,dict]]]=None,
                              fields_update:Union[List[Union[CkanField,dict]], OrderedDict[str,Union[CkanField,dict]]]=None, *,
                              only_if_needed:bool=False, fields:Union[List[Union[CkanField,dict]], OrderedDict[str,Union[CkanField,dict]]]=None,
                              fields_type_override:Dict[str,str]=None, field_description:Dict[str,str]=None,
                              fields_label:Dict[str,str]=None) -> Tuple[bool, List[dict], Union[dict,bool,None]]:
        """
        Function helper call to API datastore_create in order to update the parameters of some fields. The initial field
        configuration is taken from the mapped information or requested.
        Typically, this could be used to enforce a data type on a field. In this case, it is required to resubmit the
        resource data with the API resource_patch.
        The field_update argument would be e.g. field_update={"id": {"info": {"type_override": "text"}}}
        This is equivalent to the option field_type_override={"id": "text"}

        __NB__: it is not possible to rename a field after creation through the API. To do this, the change must be done in the database.

        :param resource_id: resource id
        :param fields_update: dictionary of field id and properties to change. The update of the property dictionary is
        recursive, ensuring only the fields appearing in the update are changed.
        This field can be overridden by the values given in field_type_override, field_description, or field_label.
        :param fields_type_override: argument to simplify the edition of the info.type_override value for each field id.
        :param field_description: argument to simplify the edition of the info.notes value for each field id
        :param fields_label: argument to simplify the edition of the info.label value for each field id
        :param only_if_needed: Cancels the request if the changes do not affect the current configuration
        :return: a tuple (update_needed, fields_new, update_dict)
        """
        update_needed, fields_update = self.datastore_field_patch_dict(fields_merge=fields_merge, fields_update=fields_update,
                                                                       fields_type_override=fields_type_override,
                                                                       fields_description=field_description, fields_label=fields_label,
                                                                       datastore_merge=True, resource_id=resource_id, return_list=True)
        if update_needed or not only_if_needed:
            return update_needed, fields_update, self.datastore_create(resource_id, fields=fields_update)
        else:
            return update_needed, fields_update, None


    ## Data deletions ------------------
    def _api_datastore_delete(self, resource_id:str, *, params:dict=None,
                             force:bool=None) -> dict:
        """
        Function to delete rows an api_datastore using api_datastore_upsert.
        If no filter is given, the whole database will be erased.
        This function is private and should not be called directly.

        :param resource_id:
        :param params:
        :param force: set to True to edit a read-only resource. If not provided, this is overridden by self.default_force
        :return:
        """
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        if params is None: params = {}
        if force is None: force = self.params.default_force
        params["resource_id"] = resource_id
        params["force"] = force
        response = self._api_action_request(f"datastore_delete", method=RequestType.Post, json=params)
        if response.success:
            return response.result
        elif response.status_code == 404 and response.success_json_loads and response.error_message["__type"] == "Not Found Error":
            resource_info = self.resource_show(resource_id)  # will trigger another error if resource does not exist
            raise CkanNotFoundError(self, "DataStore", response)
        else:
            raise response.default_error(self)

    def datastore_delete_rows(self, resource_id:str, filters:dict, *, params:dict=None,
                              force:bool=None, calculate_record_count:bool=True) -> dict:
        """
        Function to delete certain rows a DataStore using _api_datastore_delete.
        The filters are mandatory here.
        If not given, the whole database would be erased. Prefer using datastore_clear for this usage.

        :see: _api_datastore_delete()
        :param resource_id:
        :param filters:
        :param params:
        :param force: set to True to edit a read-only resource. If not provided, this is overridden by self.default_force
        :param calculate_record_count:
        :return:
        """
        if params is None: params = {}
        if force is None: force = self.params.default_force
        params["filters"] = filters
        params["calculate_record_count"] = calculate_record_count
        assert_or_raise(len(filters) > 0, InvalidParameterError("filters"))
        return self._api_datastore_delete(resource_id, params=params, force=force)

    def datastore_clear(self, resource_id:str, *, error_not_found:bool=True, params:dict=None,
                        force:bool=None, bypass_admin:bool=False) -> Union[dict,None]:
        """
        Function to clear data in a DataStore using _api_datastore_delete. Requires enable_admin=True.
        This implementation adds the option error_not_found. If set to False, no error is raised if the resource is found by the datastore is not.

        :see: _api_datastore_delete()
        :param resource_id:
        :param error_not_found: if False, does not raise an exception if the resource exists but there is not datastore
        :param params:
        :param force: set to True to edit a read-only resource. If not provided, this is overridden by self.default_force
        :param bypass_admin: option to bypass check of enable_admin
        :return:
        """
        if not bypass_admin:
            assert_or_raise(self.params.enable_admin, AdminFeatureLockedError())
        if params is None: params = {}
        if force is None: force = self.params.default_force
        try:
            result = self._api_datastore_delete(resource_id, params=params, force=force)
            self.map._record_datastore_delete(resource_id)
            return result
        except CkanNotFoundError as e:
            if not error_not_found and e.object_type == "DataStore":
                msg = f"Tried to delete DataStore of existing resource_id {resource_id} but there is no DataStore."
                if self.params.verbose_request:
                    print(msg)
            else:
                raise e from e

    def _api_resource_delete(self, resource_id:str, *, params:dict=None,
                             force:bool=None, bypass_admin:bool=False) -> dict:
        """
        Function to delete a resource. This fully removes the resource, definitively. Requires enable_admin=True.

        :param resource_id:
        :param params:
        :param force: set to True to edit a read-only resource. If not provided, this is overridden by self.default_force
        :return:
        """
        if not bypass_admin:
            assert_or_raise(self.params.enable_admin, AdminFeatureLockedError())
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        if params is None: params = {}
        if force is None: force = self.params.default_force
        params["id"] = resource_id
        # params["force"] = force
        response = self._api_action_request(f"resource_delete", method=RequestType.Post, json=params)
        if response.success:
            # update map
            self.map._record_resource_delete(resource_id)
            return response.result
        else:
            raise response.default_error(self)

    def resource_delete(self, resource_id:str, *, params:dict=None,
                        force:bool=None, bypass_admin:bool=False) -> dict:
        # function alias
        return self._api_resource_delete(resource_id, params=params, force=force, bypass_admin=bypass_admin)


    ## Datastore creation ------------------
    @staticmethod
    def default_resource_view(resource_format:str) -> Tuple[str,str]:
        """
        Definition of the default resource view based on the resource format.

        :param resource_format:
        :return:
        """
        if resource_format is None:
            resource_format = "unknown"
        resource_format = resource_format.lower()
        if resource_format == "csv":
            title = "Table"
            view_type = "recline_view"  # Data Explorer
        elif resource_format in {"json", "txt", "py"}:
            title = "Text"
            view_type = "text_view"
        elif resource_format in {"png", "svg"}:
            title = "Image"
            view_type = "image_view"
        else:
            title = None
            view_type = None
        return title, view_type

    def _api_resource_view_create(self, resource_id:str, title:Union[str,List[str]]=None, *,
                                  view_type:Union[str,List[str]]=None, params:dict=None) -> List[CkanViewInfo]:
        """
        API call to resource_view_create.

        title and view_type must have same length if specified as lists.

        :param resource_id:  resource id
        :param title: Title of the resource
        :param view_type: Type of view, typically recline_view for Data Explorer
        :param params:
        :return:
        """
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        if params is None: params = {}
        if title is None or view_type is None:
            raise ValueError("title and view_type must be specified together")
        if isinstance(view_type, str):
            view_type = [view_type]
        if isinstance(title, str):
            title = [title]
        assert(len(title) == len(view_type))
        params["resource_id"] = resource_id
        view_info_list = []
        for title_selected, view_type_selected in zip(title, view_type):
            params["title"] = title_selected
            params["view_type"] = view_type_selected
            response = self._api_action_request(f"resource_view_create", method=RequestType.Post, json=params)
            if response.success:
                view_info = CkanViewInfo(response.result)
                self.map._update_view_info(view_info)
                view_info_list.append(view_info.copy())
            else:
                raise response.default_error(self)
        return view_info_list

    def resource_view_create(self, resource_id:str, title:Union[str,List[str]]=None, *,
                              view_type:Union[str,List[str]]=None, params:dict=None,
                              error_no_default_view_type:bool=False, cancel_if_exists:bool=True) -> List[CkanViewInfo]:
        """
        Encapsulation of the API resource_view_create. If no resource view is provided to create (None),
        the function looks up the default view defined in default_resource_view.
        This function also looks at the existing views and cancels the creation of those which have the same title.
        If provided as a list, title and view_type must have same length.

        :param resource_id:
        :param title:
        :param view_type:
        :param params:
        :param error_no_default_view_type:
        :param cancel_if_exists: option to cancel an existing view if it exists (based on the title)
        :return:
        """
        if title is None and view_type is None:
            resource_info = self.get_resource_info_or_request_of_id(resource_id)
            resource_format = resource_info.format
            title, view_type = self.default_resource_view(resource_format)
            if title is None:
                title = []
                view_type = []
                msg = NoDefaultView(resource_format)
                if error_no_default_view_type:
                    raise(msg)
                else:
                    warn(str(msg))
                    return []
        if isinstance(view_type, str):
            view_type = [view_type]
        if isinstance(title, str):
            title = [title]
        assert(len(title) == len(view_type))
        if cancel_if_exists:
            resource_views_by_title = {view_info.title: view_info for view_info in self.get_resource_view_list_or_request(resource_id)}
            i_rm = []
            for i, view_title in enumerate(title):
                if view_title in resource_views_by_title.keys():
                    i_rm.append(i)
            for i in reversed(i_rm):
                title.pop(i)
                view_type.pop(i)
        if len(title) > 0:
            return self._api_resource_view_create(resource_id, title, view_type=view_type, params=params)
        else:
            return []

    def _api_resource_create(self, package_id:str, name:str, *, format:str=None, description:str=None,
                             state:CkanState=None,
                             df:pd.DataFrame=None, file_path:str=None, url:str=None, files=None,
                             payload:Union[bytes, io.BufferedIOBase]=None, payload_name:str=None,
                             params:dict=None) -> CkanResourceInfo:
        """
        API call to resource_create.

        :see: _api_resource_patch
        :see: resource_create
        :param package_id:
        :param name:
        :param format:
        :param url: url of the resource to replace resource
        :param params: additional parameters such as resource_type can be set

        For file uploads, the following parameters are taken, by order of priority:
        See upload_prepare_requests_files_arg for an example of formatting.

        :param files: files pass through argument to the requests.post function. Use to send other data formats.
        :param payload: bytes to upload as a file
        :param payload_name: name of the payload to use (associated with the payload argument) - this determines the format recognized in CKAN viewers.
        :param file_path: path of the file to transmit (binary and text files are supported here)
        :param df: pandas DataFrame to replace resource

        :return:
        """
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        if params is None: params = {}
        params["package_id"] = package_id
        params["name"] = name
        if format is not None:
            params["format"] = format
        if state is not None:
            params["state"] = str(state)
        if description is not None:
            params["description"] = description
        files = upload_prepare_requests_files_arg(files=files, file_path=file_path, df=df, payload=payload, payload_name=payload_name)
        if url is not None:
            params["url"] = url
            assert(files is None)
        if files is not None:
            response = self._api_action_request(f"resource_create", method=RequestType.Post,
                                                files=files, data=params)
        else:
            response = self._api_action_request(f"resource_create", method=RequestType.Post, json=params)
        if response.success:
            # update map
            resource_info = CkanResourceInfo(response.result)
            self.map._record_resource_create(resource_info)
            return resource_info.copy()
        else:
            raise response.default_error(self)

    def datastore_default_alias(self, resource_name:str, package_name:str, *,
                                query_names:bool=True, error_not_found:bool=True) -> str:
        if query_names:
            package_info = self.get_package_info_or_request(package_name)
            resource_info = self.get_resource_info_or_request(resource_name, package_name, error_not_found=error_not_found)
            if resource_info is not None:
                return CkanApiManage.datastore_default_alias_of_info(resource_info, package_info)
        return CkanApiManage.datastore_default_alias_of_names(resource_name, package_name)

    @staticmethod
    def datastore_default_alias_of_info(resource_info:CkanResourceInfo, package_info:CkanPackageInfo) -> str:
        package_name= package_info.name
        resource_name = resource_info.name
        return CkanApiManage.datastore_default_alias_of_names(resource_name, package_name)

    @staticmethod
    def datastore_default_alias_of_names(resource_name:str, package_name:str) -> str:
        resource_varname = clean_table_name(resource_name.lower().strip())
        alias_name = package_name + default_alias_package_resource_sep + resource_varname
        if alias_name_max_len is not None and len(alias_name) > alias_name_max_len:
            alias_hash = hashlib.sha1(alias_name.encode("utf-8")).hexdigest()
            if default_alias_hash_replace:
                return "alias" + default_alias_hash_sep + alias_hash
            else:
                alias_name_truncated = alias_name[:alias_name_max_len-default_alias_hash_len-len(default_alias_hash_sep)] + default_alias_hash_sep + alias_hash[:default_alias_hash_len]
                return alias_name_truncated
        else:
            return alias_name

    def resource_create(self, package_id:str, name:str, *, format:str=None, description:str=None, state:CkanState=None,
                        params:dict=None,
                        url:str=None,
                        files=None, file_path:str=None, df:pd.DataFrame=None,
                        payload:Union[bytes, io.BufferedIOBase]=None, payload_name:str=None,
                        cancel_if_exists:bool=True, update_if_exists:bool=False, reupload:bool=False, create_default_view:bool=True, auto_submit:bool=False,
                        datastore_create:bool=False, records:Union[dict, List[dict], pd.DataFrame]=None, fields:List[dict]=None,
                            primary_key: Union[str, List[str]] = None, indexes: Union[str, List[str]] = None,
                            aliases: Union[str, List[str]] = None, data_cleaner:CkanDataCleanerABC=None) -> CkanResourceInfo:
        """
        Proxy to API call resource_create verifying if a resource with the same name already exists and adding the default view.

        :param package_id:
        :param name:
        :param format:
        :param params:
        :param cancel_if_exists: check if a resource with the same name already exists in the package on CKAN server
        If a resource with the same name already exists, the info for this resource is returned
        :param update_if_exists: If a resource with the same name already exists (and cancel_if_exists=True), a call to resource_patch is performed.
        :param reupload: re-upload the resource if a resource with the same name already exists and cancel_if_exists=True and update_if_exists=True
        :param create_default_view:

        For file uploads, the following parameters are taken, by order of priority:
        See upload_prepare_requests_files_arg for an example of formatting.

        :param files: files pass through argument to the requests.post function. Use to send other data formats.
        :param payload: bytes to upload as a file
        :param payload_name: name of the payload to use (associated with the payload argument) - this determines the format recognized in CKAN viewers.
        :param file_path: path of the file to transmit (binary and text files are supported here)
        :param df: pandas DataFrame to replace resource

        :return:
        """
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        has_file_data = (files is not None or df is not None or file_path is not None or payload is not None)
        has_records = records is not None
        delete_previous_datastore = (has_file_data or has_records) and reupload and datastore_create
        if name is None or name == "":
            raise CkanMandatoryArgumentError("resource_create", "name")
        if df is not None and datastore_id_col in df.columns:
            msg = f"You cannot initiate a resource with an {datastore_id_col} column if you are creating a DataStore. You risk to create a conflict error."
            warn(msg)
        if cancel_if_exists:
            self.map_resources(package_id, only_missing=True)
            if name in self.map.packages[package_id].resources_id_index.keys():
                resource_info = self.map.get_resource_info(name, package_id)
                resource_id = resource_info.id
                resource_info.newly_created = False
                resource_info.newly_updated = False
                delete_previous_datastore = delete_previous_datastore # and self.resource_is_datastore(resource_id)
                if update_if_exists:
                    if has_file_data and not reupload:
                        # cancel reupload if not enabled
                        has_file_data = False
                        df = None
                        file_path = None
                        files = None
                        payload = None
                    if has_records and not reupload:
                        records = None
                        has_records = False
                    if delete_previous_datastore:
                        # if there already was a datastore, clear it
                        self.datastore_clear(resource_id, error_not_found=False, bypass_admin=True)
                    resource_info.update(self.resource_patch(resource_id, name=name, format=format, state=state,
                                                             description=description, url=url,
                                                             df=df, file_path=file_path, files=files,
                                                             payload=payload, payload_name=payload_name))
                    if create_default_view:
                        view_info_list = self.resource_view_create(resource_info.id)
                        resource_info.update_view(view_info_list)
                    if has_file_data:
                        resource_info.newly_updated = True
                        if auto_submit:
                            self.datastore_submit(resource_info.id)
                    if datastore_create or delete_previous_datastore:
                        info = self.datastore_create(resource_info.id, records=records, fields=fields, primary_key=primary_key,
                                                     indexes=indexes, aliases=aliases, delete_previous=False, data_cleaner=data_cleaner)
                    return resource_info
                else:
                    return resource_info
        # here: the resource does not exist => create a new one
        resource_info = self._api_resource_create(package_id, name, format=format, description=description, state=state,
                                                  url=url,
                                                  files=files, file_path=file_path, df=df,
                                                  payload=payload, payload_name=payload_name, params=params)
        resource_info.newly_created = True
        resource_info.newly_updated = False
        if create_default_view:
            view_info_list = self.resource_view_create(resource_info.id)
            resource_info.update_view(view_info_list)
        if auto_submit and has_file_data:
            self.datastore_submit(resource_info.id)
        if datastore_create:
            info = self.datastore_create(resource_info.id, records=records, fields=fields, primary_key=primary_key,
                                         indexes=indexes, aliases=aliases, delete_previous=False, data_cleaner=data_cleaner)
        return resource_info

    def _api_datastore_create(self, resource_id:str, *, records:Union[dict, List[dict], pd.DataFrame]=None,
                             fields:List[Union[dict, CkanField]]=None,
                             primary_key:Union[str, List[str]]=None, indexes:Union[str, List[str]]=None,
                             aliases: Union[str, List[str]]=None,
                             params:dict=None,force:bool=None) -> dict:
        """
        API call to datastore_create.
        This endpoint also supports altering tables, aliases and indexes and bulk insertion.

        :param resource_id: resource id
        :param records:
        :param fields:
        :param primary_key:
        :param indexes:
        :param params:
        :param force:
        :return:
        """
        # assert_or_raise(self.enable_admin, AdvancedFeatureLockedError())
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        if params is None: params = {}
        if force is None: force = self.params.default_force
        params["resource_id"] = resource_id
        params["force"] = force
        if primary_key is not None:
            if isinstance(primary_key, str):
                primary_key = primary_key.split(ckan_tags_sep)
            params["primary_key"] = primary_key
        if indexes is not None:
            if isinstance(indexes, str):
                indexes = indexes.split(ckan_tags_sep)
            params["indexes"] = indexes
        if aliases is not None:
            if isinstance(aliases, str):
                aliases = aliases.split(ckan_tags_sep)
            params["aliases"] = aliases
        if records is not None:
            if isinstance(records, pd.DataFrame):
                params["records"] = records.to_dict(orient='records')
            else:
                params["records"] = records
        if fields is not None:
            # list of dicts
            fields_list_dict = [field_info.to_ckan_dict() if isinstance(field_info, CkanField) else field_info for field_info in fields]
            params["fields"] = fields_list_dict
        data_payload, json_headers = json_encode_params(params)
        response = self._api_action_request(f"datastore_create", method=RequestType.Post,
                                            data=data_payload, headers=json_headers)
        # response = self._api_action_request(f"datastore_create", method=RequestType.Post, json=params)
        if response.success:
            return response.result
        else:
            raise response.default_error(self)

    def datastore_create(self, resource_id:str, *, delete_previous:bool=False, bypass_admin:bool=False,
                         records:Union[dict, List[dict], pd.DataFrame]=None,
                         fields:List[Union[dict,CkanField]]=None,
                         primary_key:Union[str, List[str]]=None, indexes:Union[str, List[str]]=None,
                         aliases: Union[str, List[str]]=None,
                         params:dict=None,force:bool=None, data_cleaner:CkanDataCleanerABC=None) -> dict:
        """
        Encapsulation of the datastore_create API call.
        This function can optionally clear the DataStore before creating it.

        :param resource_id:
        :param delete_previous: option to delete the previous datastore, if exists (default:False)
        :param records:
        :param fields:
        :param primary_key:
        :param indexes:
        :param params:
        :param force:
        :return:
        """
        if delete_previous:
            self.datastore_clear(resource_id, error_not_found=False, bypass_admin=bypass_admin)
        if data_cleaner is None:
            data_cleaner = self.data_cleaner_upload
        if data_cleaner is not None:
            if not delete_previous:
                fields_for_cleaner_dict = CkanApiManage.datastore_field_dict(fields=fields)
                fields_for_cleaner = OrderedDict([(field_name, CkanField.from_ckan_dict(field_dict)) for field_name, field_dict in fields_for_cleaner_dict.items()])
            else:
                fields_for_cleaner = None
            records = data_cleaner.clean_records(records, known_fields=fields_for_cleaner, inplace=True)
            if len(records) > 0:
                if primary_key is None:
                    primary_key = data_cleaner.field_suggested_primary_key
                if indexes is None:
                    indexes = list(data_cleaner.field_suggested_index)
                fields = data_cleaner.merge_field_changes(fields)
        if self.params.default_alias_enforce:
            resource_info = self.get_resource_info_or_request_of_id(resource_id)
            package_info = self.get_package_info_or_request(resource_info.package_id)
            if aliases is None and not delete_previous:
                datastore_info = self.get_datastore_info_or_request_of_id(resource_id, error_not_found=False)
                if datastore_info is not None:
                    aliases = datastore_info.aliases  # when aliases argument is None, the aliases are not modified => keep existing aliases from server
            default_alias_name = self.datastore_default_alias_of_info(resource_info, package_info)
            if aliases is None:
                aliases = [default_alias_name]
            else:
                aliases.append(default_alias_name)
            aliases = list(set(aliases))  # keep unique values
        return self._api_datastore_create(resource_id, records=records, fields=fields,
                                          primary_key=primary_key, indexes=indexes, aliases=aliases,
                                          params=params, force=force)


    ## Package creation/deletion/edit ------------------
    def _api_package_patch(self, package_id: str, package_name:str=None, private:bool=None, *, title:str=None, notes:str=None, owner_org:str=None,
                           state:Union[CkanState,str]=None, license_id:str=None, tags:List[str]=None, tags_list_dict:List[Dict[str, str]]=None,
                           url:str=None, version:str=None, custom_fields:dict=None,
                           author:str=None, author_email:str=None, maintainer:str=None, maintainer_email:str=None,
                           params:dict=None) -> CkanPackageInfo:
        """
        API call to package_patch. Use to change the properties of a package.
        This method is preferred to package_update which requires to resend the full package configuration.
        (API doc for package_update: It is recommended to call ckanapi_harvesters.logic.action.get.package_show(),
        make the desired changes to the result, and then call package_update() with it.)

        :param package_id:
        :param package_name:
        :param private:
        :param title:
        :param notes:
        :param owner_org:
        :param state:
        :param license_id:
        :param params:
        :return:
        """
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        if params is None: params = {}
        params["id"] = package_id
        params["private"] = private
        if owner_org is None and use_ckan_owner_org_as_default:
            owner_org = self.owner_org
        if owner_org is not None:
            params["owner_org"] = owner_org
        if package_name is not None:
            assert (2 <= len(package_name) <= 100 and re.match(ckan_package_name_re, package_name))
            params["name"] = package_name
        if title is not None:
            params["title"] = title
        if notes is not None:
            params["notes"] = notes
        if url is not None:
            params["url"] = url
        if version is not None:
            params["version"] = version
        if tags is not None:
            if tags_list_dict is None:
                tags_list_dict = []
            for tag in tags:
                tags_list_dict.append({"name": tag})
        if tags_list_dict is not None:
            params["tags"] = tags_list_dict
        if custom_fields is not None:
            params["extras"] = [{"key": key, "value": value if value is not None else ""} for key, value in custom_fields.items()]
        if author is not None:
            params["author"] = author
        if author_email is not None:
            params["author_email"] = author_email
        if maintainer is not None:
            params["maintainer"] = maintainer
        if maintainer_email is not None:
            params["maintainer_email"] = maintainer_email
        if state is not None:
            if isinstance(state, str):
                params["state"] = state
            else:
                params["state"] = str(state)
        if license_id is not None:
            params["license_id"] = license_id
        response = self._api_action_request(f"package_patch", method=RequestType.Post, json=params)
        if response.success:
            # update map
            pkg_info = CkanPackageInfo(response.result)
            self.map._record_package_update(pkg_info)
            return pkg_info.copy()
        else:
            raise response.default_error(self)

    def package_patch(self, package_id: str, package_name:str=None, private:bool=None, *, title:str=None, notes:str=None, owner_org:str=None,
                           state:Union[CkanState,str]=None, license_id:str=None, tags:List[str]=None, tags_list_dict:List[Dict[str, str]]=None,
                           url:str=None, version:str=None, custom_fields:dict=None,
                           author:str=None, author_email:str=None, maintainer:str=None, maintainer_email:str=None,
                           params:dict=None) -> CkanPackageInfo:
        # function alias
        return self._api_package_patch(package_id=package_id, package_name=package_name, private=private,
                                       title=title, notes=notes, owner_org=owner_org, state=state,
                                       license_id=license_id, tags=tags, tags_list_dict=tags_list_dict, url=url, version=version,
                                       custom_fields=custom_fields, author=author, author_email=author_email,
                                       maintainer=maintainer, maintainer_email=maintainer_email,
                                       params=params)

    def package_state_change(self, package_id:str, state:CkanState) -> CkanPackageInfo:
        """
        Change package state using the package_patch API.

        :param package_id:
        :param state:
        :return:
        """
        return self.package_patch(package_id, state=state)

    def _api_package_create(self, name:str, private:bool, *, title:str=None, notes:str=None, owner_org:str=None,
                            state: Union[CkanState, str] = None, license_id: str = None, tags: List[str] = None, tags_list_dict:List[Dict[str, str]]=None,
                            url: str = None, version: str = None, custom_fields: dict = None,
                            author: str = None, author_email: str = None,
                            maintainer: str = None, maintainer_email: str = None,
                            params:dict=None) -> CkanPackageInfo:
        """
        API call to package_create.

        :param name:
        :param private:
        :param title:
        :param notes:
        :param owner_org:
        :param state:
        :param license_id:
        :param tags:
        :param params:
        :return:
        """
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        if params is None: params = {}
        assert(2 <= len(name) <= 100 and re.match(ckan_package_name_re, name))
        params["name"] = name
        params["private"] = private
        if owner_org is None and use_ckan_owner_org_as_default:
            owner_org = self.owner_org
        if owner_org is not None:
            params["owner_org"] = owner_org
        if title is not None:
            params["title"] = title
        if notes is not None:
            params["notes"] = notes
        if url is not None:
            params["url"] = url
        if version is not None:
            params["version"] = version
        if tags is not None:
            if tags_list_dict is None:
                tags_list_dict = []
            for tag in tags:
                tags_list_dict.append({"name": tag})
        if tags_list_dict is not None:
            params["tags"] = tags_list_dict
        if custom_fields is not None:
            params["extras"] = [{"key": key, "value": value if value is not None else ""} for key, value in custom_fields.items()]
        if author is not None:
            params["author"] = author
        if author_email is not None:
            params["author_email"] = author_email
        if maintainer is not None:
            params["maintainer"] = maintainer
        if maintainer_email is not None:
            params["maintainer_email"] = maintainer_email
        if state is not None:
            if isinstance(state, str):
                params["state"] = state
            else:
                params["state"] = str(state)
        if license_id is not None:
            params["license_id"] = license_id
        response = self._api_action_request(f"package_create", method=RequestType.Post, json=params)
        if response.success:
            # update map
            pkg_info = CkanPackageInfo(response.result)
            self.map._record_package_create(pkg_info)
            return pkg_info.copy()
        else:
            raise response.default_error(self)

    def package_create(self, package_name:str, private:bool=True, *, title:str=None, notes:str=None, owner_org:str=None,
                       state: Union[CkanState, str] = None, license_id: str = None, tags: List[str] = None, tags_list_dict:List[Dict[str, str]]=None,
                       url: str = None, version: str = None, custom_fields: dict = None,
                       author: str = None, author_email: str = None,
                       maintainer: str = None, maintainer_email: str = None,
                       params:dict=None, cancel_if_exists:bool=True, update_if_exists=True) -> CkanPackageInfo:
        """
        Helper function to create a new package. This first checks if the package already exists.

        :see: _api_package_create()
        :param package_name:
        :param private:
        :param title:
        :param notes:
        :param owner_org:
        :param license_id:
        :param state:
        :param params:
        :param cancel_if_exists:
        :param update_if_exists:
        :return:
        """
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        if package_name is None or package_name == "":
            raise CkanMandatoryArgumentError("package_create", "package_name")
        self.map_resources(package_name, only_missing=True, error_not_found=False)
        pkg_info = self.map.get_package_info(package_name, error_not_mapped=False)
        if pkg_info is not None and cancel_if_exists:
            if update_if_exists:
                pkg_info = self.package_patch(pkg_info.id, package_name, private=private, title=title, notes=notes,
                                              owner_org=owner_org, state=state, license_id=license_id, tags=tags,
                                              tags_list_dict=tags_list_dict,
                                              url=url, version=version, custom_fields=custom_fields,
                                              author=author, author_email=author_email,
                                              maintainer=maintainer, maintainer_email=maintainer_email,
                                              params=params)
            pkg_info.newly_created = False
            return pkg_info
        else:
            pkg_info = self._api_package_create(package_name, private, title=title, notes=notes,
                                                owner_org=owner_org, state=state, license_id=license_id, tags=tags,
                                                tags_list_dict=tags_list_dict,
                                                url=url, version=version, custom_fields=custom_fields,
                                                author=author, author_email=author_email,
                                                maintainer=maintainer, maintainer_email=maintainer_email,
                                                params=params)
            pkg_info.newly_created = True
            return pkg_info

    def _api_package_delete(self, package_id:str,
                            *, params:dict=None) -> dict:
        """
        API call to package_delete.
        This marks the package as deleted and does not remove data.

        :param package_id:
        :param params:
        :return:
        """
        assert_or_raise(self.params.enable_admin, AdminFeatureLockedError())
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        if params is None: params = {}
        params["id"] = package_id
        response = self._api_action_request(f"package_delete", method=RequestType.Post, json=params)
        if response.success:
            # update map
            self.map._record_package_delete_state(package_id)
            return response.result
        else:
            raise response.default_error(self)

    def _api_package_resource_reorder(self, package_id:str, resource_ids: List[str],
                            *, params:dict=None) -> dict:
        """
        API call to package_resource_reorder. Reorders resources within a package.
        Reorder resources against datasets. If only partial resource ids are supplied then these are assumed to be first and the other resources will stay in their original order.

        :param package_id: the id or name of the package to update
        :param resource_ids: a list of resource ids in the order needed
        :param params:
        :return:
        """
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        if params is None: params = {}
        params["id"] = package_id
        params["order"] = resource_ids
        response = self._api_action_request(f"package_resource_reorder", method=RequestType.Post, json=params)
        if response.success:
            return response.result
        else:
            raise response.default_error(self)

    package_resource_reorder = _api_package_resource_reorder

    def _api_dataset_purge(self, package_id:str,
                           *, params:dict=None) -> dict:
        """
        API call to dataset_purge.
        This fully removes the package.
        This action is not reversible.
        It requires an admin account.

        :param package_id:
        :param params:
        :return:
        """
        assert_or_raise(self.params.enable_admin, AdminFeatureLockedError())
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        if params is None: params = {}
        params["id"] = package_id
        response = self._api_action_request(f"dataset_purge", method=RequestType.Post, json=params)
        if response.success:
            # update map
            self.map._record_package_purge_removal(package_id)
            return response.result
        else:
            raise response.default_error(self)

    def package_delete_resources(self, package_name:str, *, bypass_admin:bool=False):
        """
        Definitively delete all resources associated with the package.

        :param package_name:
        :return:
        """
        package_info = self.map.get_package_info(package_name)
        resource_ids = [resource_info.id for resource_info in package_info.package_resources.values()]
        for resource_id in resource_ids:
            self.resource_delete(resource_id, bypass_admin=bypass_admin)

    def package_delete(self, package_id:str, definitive_delete:bool=False, *, params:dict=None) -> dict:
        """
        Alias function for package removal. Either calls API package_delete to simply mark for deletion or dataset_purge
        to definitively delete the package.

        :param package_id:
        :param definitive_delete: True: calls dataset_purge (action not reversible), False: calls API package_delete.
        :param params:
        :return:
        """
        if definitive_delete:
            return self._api_dataset_purge(package_id, params=params)
        else:
            return self._api_package_delete(package_id, params=params)





#!python3
# -*- coding: utf-8 -*-
"""
Errors specific to harvesting data
"""

from ckanapi_harvesters.auxiliary.ckan_errors import RequirementError
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_errors import CleanError, CleanerRequirementError  # alias


class HarvestMethodRequiredError(Exception):
    def __init__(self):
        super().__init__("The harvesting method argument --harvester is required.")

class HarvesterArgumentError(Exception):
    pass

class HarvesterArgumentRequiredError(HarvesterArgumentError):
    def __init__(self, argument:str, harvest_method:str, help:str=None):
        if help is None: help = ""
        super().__init__(f"The argument {argument} is required for harvest method {harvest_method}. " + help)

class HarvesterRequirementError(RequirementError):
    def __init__(self, requirement:str, harvest_method:str):
        super().__init__(f"The package {requirement} is required for this harvester ({harvest_method}).")

class ResourceNotFoundError(Exception):
    def __init__(self, resource_type:str, table_name:str, host:str):
        super().__init__(f"{resource_type} {table_name} was not found on host ({host}).")





#!python3
# -*- coding: utf-8 -*-
"""
Harvester base class
"""
from typing import Union, List
from collections import OrderedDict
import copy

from ckanapi_harvesters.auxiliary.ckan_auxiliary import CkanFieldInternalAttrs


class FieldMetadata:
    def __init__(self):
        self.name:str = ""
        self.description:Union[str,None] = None
        self.label:Union[str,None] = None
        self.data_type:Union[str,None] = None
        self.is_index:Union[bool,None] = None
        self.uniquekey:Union[bool,None] = None
        self.notnull:Union[bool,None] = None
        self.internal_attrs: CkanFieldInternalAttrs = CkanFieldInternalAttrs()
        self.harvester_attrs: dict = {}

    def copy(self):
        return copy.deepcopy(self)


class TableMetadata:
    def __init__(self):
        self.name: str = ""
        self.primary_key: Union[List[str],None] = None
        self.indexes: Union[List[str],None] = None
        self.unique_keys: Union[List[str],None] = None
        self.description: Union[str,None] = None
        self.fields: Union[OrderedDict[str,FieldMetadata],None] = None

    def copy(self):
        return copy.deepcopy(self)


class DatasetMetadata:
    def __init__(self):
        self.name:str = ""
        self.description:Union[str,None] = None
        self.tables:Union[OrderedDict[str,TableMetadata],None] = None

    def copy(self):
        return copy.deepcopy(self)



#!python3
# -*- coding: utf-8 -*-
"""
Functions to define messages with an error level
"""
from enum import IntEnum
from collections import OrderedDict
from warnings import warn


class ErrorLevel(IntEnum):
    Information = 0
    Warning = 1
    Error = 2

    def __str__(self):
        return self.name.lower()

    @staticmethod
    def from_str(s):
        s = s.lower().strip()
        if s == "information":
            return ErrorLevel.Information
        elif s == "warning":
            return ErrorLevel.Warning
        elif s == "error":
            return ErrorLevel.Error
        else:
            raise ValueError(s)


class ErrorLevelMessage(Exception):
    def __init__(self, error_level:ErrorLevel, message: str):
        super().__init__(message)
        self.error_level: ErrorLevel = error_level
        self.message: str = message

    def to_dict(self) -> dict:
        return OrderedDict([
            ("level", str(self.error_level)),
            ("message", self.message),
        ])


class ContextErrorLevelMessage(ErrorLevelMessage):
    def __init__(self, context:str, error_level:ErrorLevel, specific_message: str):
        message = f"In {context} / {error_level.name}: {specific_message}"
        super().__init__(error_level, message)
        self.context: str = context
        self.specific_message: str = specific_message




#!python3
# -*- coding: utf-8 -*-
"""
Data model to represent a CKAN database architecture
"""
from abc import ABC, abstractmethod
from typing import Dict, List, Union
from warnings import warn
import copy

from ckanapi_harvesters.auxiliary.ckan_model import CkanPackageInfo, CkanResourceInfo, CkanState, CkanDataStoreInfo, \
    CkanOrganizationInfo, CkanLicenseInfo, CkanViewInfo, CkanGroupInfo, CkanUserInfo
from ckanapi_harvesters.auxiliary.ckan_errors import NotMappedObjectNameError, IntegrityError
from ckanapi_harvesters.auxiliary.ckan_auxiliary import assert_or_raise


class CkanMapABC(ABC):
    @abstractmethod
    def purge(self):
        raise NotImplementedError()

    @abstractmethod
    def copy(self):
        raise NotImplementedError()

    @abstractmethod
    def to_dict(self) -> dict:
        raise NotImplementedError()

    @abstractmethod
    def update_from_dict(self, data:dict) -> None:
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def from_dict(d: dict) -> "CkanMap":
        raise NotImplementedError()


class CkanMap(CkanMapABC):
    """
    Class to store an image of the CKAN database architecture
    Auxiliary class of CkanApi
    """

    def __init__(self):
        self.packages:Dict[str,CkanPackageInfo] = {}    # package id -> info
        self.packages_id_index:Dict[str, str] = {}      # package name -> id
        self.packages_title_index:Dict[str, str] = {}   # package title -> id
        self.resources:Dict[str,CkanResourceInfo] = {}  # resource id -> info
        self.resource_alias_index:Dict[str,str] = {}    # resource alias -> id
        self.licenses:Dict[str,CkanLicenseInfo] = {}    # license id -> info
        self.licenses_title_index:Dict[str, str] = {}   # license title -> id
        self.organizations:Dict[str,CkanOrganizationInfo] = {} # organization id -> info
        self.organizations_id_index:Dict[str, str] = {} # organization name -> id
        self.organizations_title_index:Dict[str, str] = {} # organization title -> id
        self.users:Dict[str,CkanUserInfo] = {}          # user id -> info
        self.users_id_index:Dict[str, str] = {}         # user name -> id
        self.groups:Dict[str,CkanGroupInfo] = {}        # group id -> info
        self.groups_id_index:Dict[str, str] = {}        # group name -> id
        self.groups_title_index:Dict[str, str] = {}     # group title -> id
        self.organizations_listed_all:bool = False
        self.organizations_listed_all_users:bool = False
        self.users_listed_all:bool = False
        self.groups_listed_all:bool = False
        self._mapping_query_datastore_info = False      # default request for datastore_info during mapping operations
        self._mapping_query_resource_view_list = False  # False: do not request resource_view_list by default
        self._mapping_query_license_list = False        # False by default
        self._mapping_query_organization_info = False   # False by default

    def purge(self):
        """
        Erase known package mappings.

        :return:
        """
        self.packages:Dict[str,CkanPackageInfo] = {}    # package id -> info
        self.packages_id_index:Dict[str, str] = {}      # package name -> id
        self.packages_title_index:Dict[str, str] = {}   # package title -> id
        self.resources:Dict[str,CkanResourceInfo] = {}  # resource id -> info
        self.resource_alias_index:Dict[str,str] = {}    # resource alias -> id
        self.licenses:Dict[str,CkanLicenseInfo] = {}    # license id -> info
        self.licenses_title_index:Dict[str, str] = {}      # license title -> id
        self.organizations:Dict[str,CkanOrganizationInfo] = {} # organization id -> info
        self.organizations_id_index:Dict[str, str] = {} # organization name -> id
        self.organizations_title_index:Dict[str, str] = {} # organization title -> id
        self.organizations_listed_all = False

    def copy(self) -> "CkanMap":
        return copy.deepcopy(self)

    def to_dict(self) -> dict:
        return {"packages":[package.to_dict() for package in self.packages.values()],
                "licenses":[license.to_dict() for license in self.licenses.values()],
                "organizations":[organization.to_dict() for organization in self.organizations.values()],
                }

    def update_from_dict(self, data:dict) -> None:
        for package_dict in data["packages"]:
            self._update_package_info(CkanPackageInfo.from_dict(package_dict))
        for license_dict in data["licenses"]:
            self._update_license_info(CkanLicenseInfo.from_dict(license_dict))
        for org_dict in data["organizations"]:
            self._update_organization_info(CkanOrganizationInfo.from_dict(org_dict))

    @staticmethod
    def from_dict(d: dict) -> "CkanMap":
        map = CkanMap()
        map.update_from_dict(d)
        return map

    ## Resource ID Map navigation  ------------------
    def get_package_id(self, package_name:str, *, error_not_mapped:bool=True, search_title:bool=True) -> Union[str,None]:
        """
        Retrieve the package id for a given package name based on the package map.

        :param package_name: package name or id.
        :return:
        """
        if package_name is None:
            raise ValueError("package_name cannot be None")
        if package_name in self.packages.keys():
            # recognized package_id
            package_id = package_name
        elif package_name in self.packages_id_index.keys():
            package_id = self.packages_id_index[package_name]
        elif search_title and package_name in self.packages_title_index.keys():
            package_id = self.packages_title_index[package_name]
        elif error_not_mapped:
            raise NotMappedObjectNameError(f"Package {package_name} is not mapped or does not exist.")
        else:
            package_id = None
        return package_id

    def get_resource_id(self, resource_name:str, package_name:str=None, *, error_not_mapped:bool=True) -> Union[str,None]:
        """
        Retrieve the resource id for a given combination of (package name and resource name) based on the package map.

        :param resource_name: resource alias, name or id.
        :param package_name: package name or id (required if resource_name is a resource name). An integrity check is performed if given.
        :return:
        """
        if resource_name in self.resources.keys():
            # recognized resource_id
            resource_id = resource_name
        elif resource_name in self.resource_alias_index.keys():
            # found resource alias
            resource_id = self.resource_alias_index[resource_name]
        else:
            package_id = self.get_package_id(package_name, error_not_mapped=error_not_mapped)
            if package_id is None and not error_not_mapped:
                return None
            if resource_name in self.packages[package_id].resources_id_index.keys():
                resource_id = self.packages[package_id].resources_id_index[resource_name]
            elif error_not_mapped:
                raise NotMappedObjectNameError(f"Resource {resource_name} is not mapped or does not exist.")
            else:
                resource_id = None
        # sanity check
        if package_name is not None and resource_id is not None:
            resource_info = self.resources[resource_id]
            map_package_id = self.get_package_id(package_name, error_not_mapped=error_not_mapped)
            if map_package_id is not None:
                assert_or_raise(map_package_id == resource_info.package_id, IntegrityError("package_id"))
        return resource_id

    def get_organization_id(self, organization_name:str, *, error_not_mapped:bool=True, search_title:bool=True) -> Union[str,None]:
        """
        Retrieve the organization id for a given organization name based on the mapped data.

        :param organization_name: organization name, title or id.
        :return:
        """
        if organization_name is None:
            raise ValueError("organization_name cannot be None")
        if organization_name in self.organizations.keys():
            # recognized organization_id
            organization_id = organization_name
        elif organization_name in self.organizations_id_index.keys():
            organization_id = self.organizations_id_index[organization_name]
        elif search_title and organization_name in self.organizations_title_index.keys():
            organization_id = self.organizations_title_index[organization_name]
        elif error_not_mapped:
            raise NotMappedObjectNameError(f"Organization {organization_name} is not mapped or does not exist.")
        else:
            organization_id = None
        return organization_id

    def get_resource_info(self, resource_name:str, package_name:str=None, *, error_not_mapped:bool=True) -> Union[CkanResourceInfo,None]:
        """
        Retrieve the information on a given resource.

        :param resource_name: resource name or id.
        :param package_name: package name or id (required if resource_name is a resource name). An integrity check is performed if given.
        :return:
        """
        resource_id = self.get_resource_id(resource_name, package_name, error_not_mapped=error_not_mapped)
        if resource_id is not None:
            return self.resources[resource_id]
        else:
            return None

    def get_package_info(self, package_name:str, *, error_not_mapped:bool=True) -> Union[CkanPackageInfo,None]:
        """
        Retrieve the package info for a given package name based on the package map.

        :param package_name: package name or id.
        :return:
        """
        package_id = self.get_package_id(package_name, error_not_mapped=error_not_mapped)
        if package_id is not None:
            return self.packages[package_id]
        else:
            return None

    def get_organization_info(self, organization_name:str, *, error_not_mapped:bool=True) -> Union[CkanOrganizationInfo,None]:
        """
        Retrieve the organization info for a given organization name based on the mapped data.

        :param organization_name: organization name or id.
        :return:
        """
        organization_id = self.get_organization_id(organization_name, error_not_mapped=error_not_mapped)
        if organization_id is not None:
            return self.organizations[organization_id]
        else:
            return None

    def get_organization_for_owner_org(self, organization_name:str, *, error_not_mapped:bool=True) -> Union[CkanOrganizationInfo,None]:
        """
        Retrieve the organization name for a given organization name based on the mapped data.
        This is the field usually used for the owner_org argument. Calls CkanOrganizationInfo.get_owner_org

        :param organization_name: organization name or id.
        :return:
        """
        organization_info = self.get_organization_info(organization_name, error_not_mapped=error_not_mapped)
        if organization_info is not None:
            return organization_info.get_owner_org()
        else:
            return None

    def get_resource_package_id(self, resource_name:str, package_name:str=None, *, error_not_mapped:bool=True) -> Union[str,None]:
        """
        Retrieve the package id of a given resource.

        :param resource_name: resource name or id.
        :param package_name: package name or id (required if resource_name is a resource name). An integrity check is performed if given.
        :return:
        """
        resource_info = self.get_resource_info(resource_name, package_name, error_not_mapped=error_not_mapped)
        if resource_info is not None:
            return resource_info.package_id
        else:
            return None

    def get_datastore_info(self, resource_name:str, package_name:str=None, *, error_not_mapped:bool=True) -> Union[CkanDataStoreInfo,None]:
        """
        :param resource_name: resource name or id.
        :param package_name: package name or id (required if resource_name is a resource name). An integrity check is performed if given.
        :return:
        """
        resource_info = self.get_resource_info(resource_name, package_name, error_not_mapped=error_not_mapped)
        if resource_info is not None:
            if resource_info.datastore_info is not None:
                return resource_info.datastore_info
            elif error_not_mapped:
                raise NotMappedObjectNameError(f"DataStore of resource {resource_name} is not mapped or does not exist.")
            else:
                return None
        else:
            return None

    def get_datastore_len(self, resource_name:str, package_name:str=None, *, error_not_mapped:bool=True) -> Union[int,None]:
        """
        Retrieve the number of rows in a DataStore from the mapped data. This requires the map_resources to be called with the option datastore_info=True.

        :param resource_name: resource name or id.
        :param package_name: package name or id (required if resource_name is a resource name). An integrity check is performed if given.
        :return:
        """
        datastore_info = self.get_datastore_info(resource_name, package_name, error_not_mapped=error_not_mapped)
        if datastore_info is not None:
            return datastore_info.row_count
        else:
            return None

    def _update_datastore_len(self, resource_id:str, new_len:int) -> None:
        """
        Internal function to update the length of a DataStore without making a request.

        :param resource_id: resource id.
        :param new_len: value to replace
        """
        resource_info = self.resources[resource_id]
        package_id = resource_info.package_id
        self.resources[resource_id].datastore_info.row_count = new_len
        self.resources[resource_id].datastore_info.details["meta"]["count"] = new_len
        self.packages[package_id].package_resources[resource_id].datastore_info.row_count = new_len
        self.packages[package_id].package_resources[resource_id].datastore_info.details["meta"]["count"] = new_len

    def _update_datastore_info(self, datastore_info:CkanDataStoreInfo) -> None:
        """
        Internal function to update the length of a DataStore without making a request.
        """
        resource_id = datastore_info.resource_id
        if resource_id in self.resources.keys():
            resource_info = self.resources[resource_id]
            package_id = resource_info.package_id
            self.resources[resource_id].datastore_info = datastore_info
            self.packages[package_id].package_resources[resource_id].datastore_info = datastore_info
            self.packages[package_id].resources_id_index[resource_info.name] = resource_id
        if datastore_info is not None and datastore_info.aliases is not None:
            self.resource_alias_index.update({alias: resource_id for alias in datastore_info.aliases})

    def _update_resource_info(self, resource_info:Union[CkanResourceInfo, List[CkanResourceInfo]]) -> None:
        """
        Internal function to update the length of a DataStore without making a request.
        """
        if not(isinstance(resource_info, list)):
            resource_info = [resource_info]
        for res_info in resource_info:
            resource_id = res_info.id
            package_id = res_info.package_id
            res_info.index_in_package = None
            if package_id in self.packages.keys():
                self.packages[package_id].update_resource(res_info)
            self.resources[resource_id] = res_info
            if res_info.datastore_info is not None and res_info.datastore_info.aliases is not None:
                self.resource_alias_index.update({alias: res_info.id for alias in res_info.datastore_info.aliases})

    def _update_view_info(self, view_info:Union[CkanViewInfo, List[CkanViewInfo]], view_list:bool=False) -> None:
        if isinstance(view_info, CkanViewInfo):
            view_info = [view_info]
        for view_info_update in view_info:
            resource_id = view_info_update.resource_id
            self.resources[resource_id].update_view(view_info_update, view_list=view_list)

    def _update_package_info(self, package_info:Union[CkanPackageInfo, List[CkanPackageInfo]]) -> None:
        """
        Internal function to update the information of a package.

        NB: the indicator pkg_info.requested_datastore_info remains False until map_resources is called.
        """
        if not(isinstance(package_info, list)):
            package_info = [package_info]
        # already done by __init__:
        # for pkg_info in package_info:
        #     pkg_info.resources_id_index.update({resource_info.name: resource_info.id for resource_info in pkg_info.resources})
        self.packages.update({pkg_info.id: pkg_info for pkg_info in package_info})
        self.packages_id_index.update({pkg_info.name: pkg_info.id for pkg_info in package_info})
        self.packages_title_index.update({pkg_info.title: pkg_info.id for pkg_info in package_info})
        for pkg_info in package_info:
            self.resources.update({resource_info.id: resource_info for resource_info in pkg_info.package_resources.values()})
            for resource_info in pkg_info.package_resources.values():
                if resource_info.datastore_info is not None and resource_info.datastore_info.aliases is not None:
                    self.resource_alias_index.update({alias: resource_info.id for alias in resource_info.datastore_info.aliases})
        for pkg_info in package_info:
            if pkg_info.organization_info is not None:
                self._update_organization_info(pkg_info.organization_info)
            if pkg_info.groups is not None:
                self._update_group_info(pkg_info.groups)


    def get_license_id(self, license_name: str, *, error_not_mapped: bool = True) -> str:
        """
        Retrieve the ID of a license based on the mapped data.

        :param license_name: license title or id.
        :return:
        """
        if license_name is None:
            raise ValueError("license_name cannot be None")
        if license_name in self.licenses.keys():
            # recognized license_id
            license_id = license_name
        elif license_name in self.licenses_title_index.keys():
            license_id = self.licenses_title_index[license_name]
        elif error_not_mapped:
            raise NotMappedObjectNameError(f"License {license_name} is not mapped or does not exist.")
        else:
            license_id = None
        return license_id

    def get_license_info(self, license_name: str, *, error_not_mapped: bool = True) -> Union[CkanLicenseInfo,None]:
        """
        Retrieve the information on a license based on the mapped data.

        :param license_name: license title or id.
        :return:
        """
        license_id = self.get_license_id(license_name, error_not_mapped=error_not_mapped)
        if license_id is not None:
            return self.licenses[license_id]
        else:
            return None

    def _update_license_info(self, license_info: Union[CkanLicenseInfo, List[CkanLicenseInfo]]) -> None:
        """
        Internal function to update the information on a license.
        """
        if not (isinstance(license_info, list)):
            license_info = [license_info]
        self.licenses.update({license.id: license for license in license_info})
        self.licenses_title_index.update({license.title: license.id for license in license_info})

    ## Package record changes  ------------------
    def _record_package_update(self, pkg_info: CkanPackageInfo) -> None:
        package_id = pkg_info.id
        package_name = pkg_info.name
        self.packages[package_id].update(pkg_info)
        self.packages_id_index[package_name] = package_id
        self.packages_title_index[pkg_info.title] = package_id

    def _record_package_create(self, pkg_info: CkanPackageInfo) -> None:
        package_id = pkg_info.id
        package_name = pkg_info.name
        self.packages[package_id] = pkg_info
        self.packages_id_index[package_name] = package_id
        self.packages_title_index[pkg_info.title] = package_id

    def _record_package_delete_state(self, package_id: str) -> None:
        # only pass in delete state
        pkg_info = self.get_package_info(package_id, error_not_mapped=False)
        if pkg_info is not None:
            pkg_info.state = CkanState.Deleted

    def _record_package_purge_removal(self, package_id:str) -> None:
        # purge = full removal
        pkg_info = self.get_package_info(package_id, error_not_mapped=False)
        if pkg_info is None:
            return
        if package_id in self.packages.keys():
            self.packages.pop(package_id)
        if pkg_info.name in self.packages_id_index.keys():
            self.packages_id_index.pop(pkg_info.name)
            self.packages_title_index.pop(pkg_info.title)

    ## Resource record changes  ------------------
    def _record_resource_update(self, resource_info:CkanResourceInfo) -> None:
        resource_id = resource_info.id
        new_resource = resource_id not in self.resources.keys()
        self.resources[resource_id] = resource_info
        if new_resource:
            self.packages[resource_info.package_id].package_resources[resource_id] = resource_info
        self.packages[resource_info.package_id].resources_id_index[resource_info.name] = resource_id

    def _record_resource_create(self, resource_info:CkanResourceInfo) -> None:
        resource_id = resource_info.id
        new_resource = resource_id not in self.resources.keys()
        self.resources[resource_id] = resource_info
        if new_resource:
            self.packages[resource_info.package_id].package_resources[resource_id] = resource_info
        self.packages[resource_info.package_id].resources_id_index[resource_info.name] = resource_id

    def _record_resource_delete(self, resource_id:str) -> None:
        if resource_id not in self.resources.keys():
            msg = f"Resource {resource_id} not found in mapped objects"
            warn(msg)
            return
        resource_info = self.resources[resource_id]
        self.resources.pop(resource_id)
        if resource_id in self.packages[resource_info.package_id].package_resources.keys():
            self.packages[resource_info.package_id].package_resources.pop(resource_id)
        if resource_info.name in self.packages[resource_info.package_id].resources_id_index.keys():
            self.packages[resource_info.package_id].resources_id_index.pop(resource_info.name)

    def _record_datastore_delete(self, resource_id:str) -> None:
        if resource_id not in self.resources.keys():
            msg = f"DataStore {resource_id} not found in mapped objects"
            warn(msg)
            return
        resource_info = self.resources[resource_id]
        resource_info.datastore_info = None
        if resource_id in self.packages[resource_info.package_id].package_resources.keys():
            self.packages[resource_info.package_id].package_resources[resource_id].datastore_info = None

    ## Organization record changes  ------------------
    def _update_organization_info(self, organization_info:Union[CkanOrganizationInfo, List[CkanOrganizationInfo]]) -> None:
        """
        Internal function to update information on an organization.
        """
        if not(isinstance(organization_info, list)):
            organization_info = [organization_info]
        self.organizations.update({info.id: info for info in organization_info})
        self.organizations_id_index.update({info.name: info.id for info in organization_info})
        self.organizations_title_index.update({info.title: info.id for info in organization_info})

    ## Group and users record changes  ------------------
    def _update_group_info(self, group_info:Union[CkanGroupInfo, List[CkanGroupInfo]]) -> None:
        """
        Internal function to update information on a group.
        """
        if not(isinstance(group_info, list)):
            group_info = [group_info]
        self.groups.update({info.id: info for info in group_info})
        self.groups_id_index.update({info.name: info.id for info in group_info})
        self.groups_title_index.update({info.title: info.id for info in group_info})

    def _update_user_info(self, user_info:Union[CkanUserInfo, List[CkanUserInfo]]) -> None:
        """
        Internal function to update information on a group.
        """
        if not(isinstance(user_info, list)):
            user_info = [user_info]
        self.users.update({info.id: info for info in user_info})
        self.users_id_index.update({info.name: info.id for info in user_info})





#!python3
# -*- coding: utf-8 -*-
"""
Code to initiate a DataStore defined by a large number of files to concatenate into one table
"""
from concurrent.futures import ThreadPoolExecutor
import threading
from threading import current_thread
from abc import ABC, abstractmethod
from typing import Dict, List, Callable, Any, Tuple, Generator, Union, Set
from warnings import warn
import copy

import pandas as pd

from ckanapi_harvesters.builder.builder_resource_datastore import BuilderDataStoreABC
from ckanapi_harvesters.builder.builder_aux import positive_end_index
from ckanapi_harvesters.auxiliary.ckan_model import UpsertChoice, CkanResourceInfo
from ckanapi_harvesters.auxiliary.ckan_auxiliary import datastore_id_col
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.builder.mapper_datastore_multi import RequestFileMapperABC, default_file_mapper_from_primary_key
from ckanapi_harvesters.builder.builder_resource_multi_file import BuilderMultiABC, default_progress_callback

# apply last_condition for each upsert request when in a multi-threaded upload on a same DataStore:
datastore_multi_threaded_always_last_condition:bool = True
# when there are multiple files, apply last insertion commands after each document? True: after each csv file, False: only at the end
datastore_multi_apply_last_condition_intermediary:bool = False


class BuilderDataStoreMultiABC(BuilderDataStoreABC, BuilderMultiABC, ABC):
    """
    generic class to manage large DataStore, divided into files/parts
    This abstract class is intended to be overloaded in order to be used to generate data from the workspace, without using CSV files
    """

    def __init__(self, *, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, dirname:str=None):
        super().__init__(name=name, format=format, description=description, resource_id=resource_id, download_url=download_url)
        # Functions inputs/outputs
        self.df_mapper: RequestFileMapperABC = default_file_mapper_from_primary_key(self.primary_key)
        self.reupload_if_needed = False  # do not reupload if needed because this could cause data loss (the upload function only uploads the first table)
        self.upsert_method: UpsertChoice = UpsertChoice.Upsert
        # BuilderMultiABC:
        self.stop_event = threading.Event()
        self.thread_ckan: Dict[str, CkanApi] = {}
        self.progress_callback: Union[Callable[[int, int, Any], None], None] = default_progress_callback
        self.progress_callback_kwargs: dict = {}
        self.enable_multi_threaded_upload:bool = True
        self.enable_multi_threaded_download:bool = True

    def copy(self, *, dest=None):
        super().copy(dest=dest)
        dest.reupload_if_needed = self.reupload_if_needed
        # BuilderMultiABC:
        dest.progress_callback = self.progress_callback
        dest.progress_callback_kwargs = copy.deepcopy(self.progress_callback_kwargs)
        dest.enable_multi_threaded_upload = self.enable_multi_threaded_upload
        dest.enable_multi_threaded_download = self.enable_multi_threaded_download
        # do not copy stop_event
        return dest

    ## upload ---------
    @abstractmethod
    def get_local_df_generator(self, resources_base_dir:str) -> Generator[pd.DataFrame, None, None]:
        """
        Returns an iterator over the parts of the upload, loaded as DataFrames (not recommended in a multi-threaded context).
        """
        raise NotImplementedError()

    @abstractmethod
    def load_local_df(self, file: Any, **kwargs) -> pd.DataFrame:
        """
        Load the DataFrame pointed by the upload part "file"
        """
        raise NotImplementedError()

    # do not change default argument apply_last_condition=True
    # def upsert_request_df(self, ckan: CkanApi, df_upload:pd.DataFrame,
    #                       method:UpsertChoice=UpsertChoice.Upsert,
    #                       apply_last_condition:bool=None) -> Tuple[pd.DataFrame, pd.DataFrame]:
    #     # calls super method, with apply_last_condition defaulting to datastore_multi_apply_last_condition_intermediary
    #     if apply_last_condition is None:
    #         apply_last_condition = True  # datastore_multi_apply_last_condition_intermediary
    #     return super().upsert_request_df(ckan=ckan, df_upload=df_upload, method=method,
    #                                      apply_last_condition=apply_last_condition)

    def _get_primary_key_indexes(self, data_cleaner_index: Set[str], current_fields:Set[str], error_missing:bool, empty_datastore:bool=False) -> Tuple[Union[List[str],None], Union[List[str],None]]:
        primary_key, indexes = super()._get_primary_key_indexes(data_cleaner_index, current_fields, error_missing, empty_datastore)
        # it is highly recommended to specify a primary key: warning if not defined
        if primary_key is None:
            msg = f"It is highly recommended to specify the primary key for a DataStore defined from a directory to ensure no duplicate values are upserted to the database. Resource: {self.name}"
            warn(msg)
        else:
            ultra_required_fields = set(primary_key)
            missing_fields = ultra_required_fields
            if current_fields is not None:
                missing_fields -= current_fields
            if len(missing_fields) > 0:
                msg = f"The primary key {self.primary_key} is set for resource {self.name} but it is not present in the sample data."
                warn(msg)
        if primary_key is None or len(primary_key) == 0:
            self.upsert_method = UpsertChoice.Insert  # do not use upsert
        return primary_key, indexes

    def upsert_request_final(self, ckan: CkanApi, *, force:bool=False) -> None:
        """
        Final steps after the last upsert query.
        This call is mandatory at the end of all requests if the user called upsert_request_df for a multi-part DataStore manually.

        :param ckan:
        :param force: perform request anyways
        :return:
        """
        force = force or not datastore_multi_apply_last_condition_intermediary
        return super().upsert_request_final(ckan, force=force)

    def upload_request_final(self, ckan: CkanApi, *, force:bool=False) -> None:
        return self.upsert_request_final(ckan=ckan, force=force)

    def upsert_request_df_no_return(self, ckan: CkanApi, df_upload:pd.DataFrame,
                                    method:UpsertChoice=UpsertChoice.Upsert,
                                    apply_last_condition:bool=None, always_last_condition:bool=None) -> None:
        """
        Calls upsert_request_df but does not return anything

        :return:
        """
        self.upsert_request_df(ckan=ckan, df_upload=df_upload, method=method,
                               apply_last_condition=apply_last_condition, always_last_condition=always_last_condition)
        return None

    def _unit_upload_apply(self, *, ckan: CkanApi, file: str,
                           index: int, start_index: int, end_index: int, total: int,
                           method: UpsertChoice, **kwargs) -> None:
        if index == 0 and self.upsert_method == UpsertChoice.Insert:
            return  # do not reupload the first document, which was used for the initialization of the dataset
        if start_index <= index and index < end_index:
            df_upload_local = self.load_local_df(file, **kwargs)
            self._call_progress_callback(index, total, info=df_upload_local,
                                         context=f"{ckan.identifier} single-thread upload")
            self.upsert_request_df_no_return(ckan=ckan, df_upload=df_upload_local, method=method,
                                             apply_last_condition=datastore_multi_apply_last_condition_intermediary)
        else:
            # self._call_progress_callback(index, total, info=df_upload_local, context=f"{ckan.identifier} single-thread skip")
            pass

    def upload_request_full(self, ckan:CkanApi, resources_base_dir:str, *,
                            method:UpsertChoice=None,
                            threads:int=1, external_stop_event=None,
                            only_missing:bool=False,
                            start_index:int=0, end_index:int=None, **kwargs) -> None:
        self.df_mapper.upsert_only_missing_rows = only_missing
        if method is None:
            if self.primary_key is None or len(self.primary_key) == 0:
                self.upsert_method = UpsertChoice.Insert  # do not use upsert if there is no primary key
            method = self.upsert_method
        super().upload_request_full(ckan=ckan, resources_base_dir=resources_base_dir,
                                    threads=threads, external_stop_event=external_stop_event,
                                    start_index=start_index, end_index=end_index,
                                    method=method, **kwargs)
        # if threads < 0:
        #     # cancel large uploads in this case
        #     return None
        # elif threads is None or threads > 1:
        #     return self.upload_request_full_multi_threaded(resources_base_dir=resources_base_dir, ckan=ckan, method=method,
        #                                                    threads=threads, external_stop_event=external_stop_event,
        #                                                    start_index=start_index, end_index=end_index)
        # else:
        #     self.init_local_files_list(resources_base_dir=resources_base_dir, cancel_if_present=True)
        #     if ckan.verbose_extra:
        #         print(f"Launching single-threaded upload of multi-file resource {self.name}")
        #     total = self.get_local_file_len()
        #     end_index = positive_end_index(end_index, total)
        #     for index, file in enumerate(self.get_local_file_generator(resources_base_dir=resources_base_dir)):
        #         if external_stop_event is not None and external_stop_event.is_set():
        #             print(f"{ckan.identifier} Interrupted")
        #             return
        #         self._unit_upload_apply(ckan=ckan, file=file,
        #                                 index=index, start_index=start_index, end_index=end_index, total=total,
        #                                 method=method)
        #     self._call_progress_callback(total, total, context=f"{ckan.identifier} single-thread upload")
        #     # at last, apply final actions:
        #     self.upload_request_final(ckan, force=not datastore_multi_apply_last_condition_intermediary)

    # def upsert_request_file_graceful(self, ckan: CkanApi, file: Any, index:int,
    #                                  method: UpsertChoice = UpsertChoice.Upsert, external_stop_event=None,
    #                                  start_index:int=0, end_index:int=None) -> None:
    #     """
    #     Calls upsert_request_df_clear with checks specific to multi-threading.
    #
    #     :return:
    #     """
    #     # ckan.session_reset()
    #     # ckan.identifier = current_thread().name
    #     ckan = self.thread_ckan[current_thread().name]
    #     total = self.get_local_file_len()
    #     end_index = positive_end_index(end_index, total)
    #     if self.stop_event.is_set():
    #         return
    #     if external_stop_event is not None and external_stop_event.is_set():
    #         print(f"{ckan.identifier} Interrupted")
    #         return
    #     try:
    #         self._unit_upload_apply(ckan=ckan, file=file,
    #                                 index=index, start_index=start_index, end_index=end_index, total=total,
    #                                 method=method)
    #     except Exception as e:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print(f"Stopping all threads because an exception occurred in thread: {e}")
    #         raise e from e

    # def upload_request_full_multi_threaded(self, ckan: CkanApi, resources_base_dir: str, threads: int = None,
    #                                        method: UpsertChoice = UpsertChoice.Upsert, external_stop_event=None,
    #                                        start_index:int=0, end_index:int=None, **kwargs):
    #     """
    #     Multi-threaded implementation of upload_request_full, using ThreadPoolExecutor.
    #     """
    #     self.init_local_files_list(resources_base_dir=resources_base_dir, cancel_if_present=True)
    #     resource_id = self.get_or_query_resource_id(ckan=ckan, error_not_found=True)  # prepare CKAN object for multi-threading: perform mapping requests if necessary
    #     self._prepare_for_multithreading(ckan)
    #     try:
    #         with ThreadPoolExecutor(max_workers=threads, initializer=self._init_thread, initargs=(ckan,)) as executor:
    #             if ckan.verbose_extra:
    #                 print(f"Launching multi-threaded upload of multi-file resource {self.name}")
    #             futures = [executor.submit(self.upsert_request_file_graceful, ckan=ckan, file=file, method=method, index=index,
    #                                        start_index=start_index, end_index=end_index, external_stop_event=external_stop_event)
    #                        for index, file in enumerate(self.get_local_file_generator(resources_base_dir=resources_base_dir))]
    #             for future in futures:
    #                 future.result()  # This will propagate the exception
    #         total = self.get_local_file_len()
    #         self._call_progress_callback(total, total, context=f"{ckan.identifier} multi-thread upload")
    #     except Exception as e:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print(f"Stopping all threads because an exception occurred: {e}")
    #         raise e from e
    #     finally:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print("End of multi-threaded upload...")
    #     # at last, apply final actions:
    #     self.upload_request_final(ckan, force=not datastore_multi_apply_last_condition_intermediary)


    ## download -------
    def download_request_df(self, ckan: CkanApi, file_query:dict) -> Union[pd.DataFrame,None]:
        """
        Download the DataFrame with the file_query arguments
        """
        resource_id = self.get_or_query_resource_id(ckan, error_not_found=self.download_error_not_found)
        if resource_id is None and not self.download_error_not_found:
            return None
        df_download = self.df_mapper.download_file_query(ckan=ckan, resource_id=resource_id, file_query=file_query)
        df = self.df_mapper.df_download_alter(df_download, file_query=file_query, fields=self._get_fields_info())
        return df

    def _unit_download_apply(self, ckan:CkanApi, file_query_item:Any, out_dir:str,
                           index:int, start_index:int, end_index:int, total:int) -> Any:
        if start_index <= index and index < end_index:
            self._call_progress_callback(index, total, info=file_query_item,
                                         context=f"{ckan.identifier} single-thread download")
            self.download_file_query_item(ckan=ckan, out_dir=out_dir, file_query_item=file_query_item)
        else:
            pass
            # self._call_progress_callback(index, total, info=file_query_item, context=f"{ckan.identifier} single-thread skip")

    def download_request_full(self, ckan: CkanApi, out_dir: str, threads:int=1, external_stop_event=None,
                              start_index:int=0, end_index:int=None, force:bool=False) -> None:
        return super().download_request_full(ckan=ckan, out_dir=out_dir,
                                             threads=threads, external_stop_event=external_stop_event,
                                             start_index=start_index, end_index=end_index, force=force)
        # if (not self.enable_download) and (not force):
        #     msg = f"Did not download resource {self.name} because download was disabled."
        #     warn(msg)
        #     return None
        # if threads < 0:
        #     # do not download large datasets in this case
        #     return None
        # elif threads is None or threads > 1:
        #     return self.download_request_full_multi_threaded(ckan=ckan, out_dir=out_dir,
        #                                                      threads=threads, external_stop_event=external_stop_event,
        #                                                      start_index=start_index, end_index=end_index)
        # else:
        #     self.init_download_file_query_list(ckan=ckan, out_dir=out_dir, cancel_if_present=True)
        #     if ckan.verbose_extra:
        #         print(f"Launching single-threaded download of multi-file resource {self.name}")
        #     total = self.get_file_query_len()
        #     end_index = positive_end_index(end_index, total)
        #     for index, file_query_item in enumerate(self.get_file_query_generator()):
        #         if external_stop_event is not None and external_stop_event.is_set():
        #             print(f"{ckan.identifier} Interrupted")
        #             return
        #         self._unit_download_apply(ckan=ckan, file_query_item=file_query_item,
        #                                   index=index, start_index=start_index, end_index=end_index, total=total)
        #     self._call_progress_callback(total, total, context=f"{ckan.identifier} single-thread download")

    def download_request_generator(self, ckan: CkanApi, out_dir: str) -> Generator[Tuple[Any, pd.DataFrame], Any, None]:
        """
        Iterator on file_queries.
        """
        self.init_download_file_query_list(ckan=ckan, out_dir=out_dir, cancel_if_present=True)
        for file_query_item in self.get_file_query_generator():
            yield self.download_file_query_item(ckan=ckan, out_dir=out_dir, file_query_item=file_query_item)

    # def download_file_query_item_graceful(self, ckan: CkanApi, out_dir: str, file_query_item: Any, index:int,
    #                                       external_stop_event=None, start_index:int=0, end_index:int=None) -> None:
    #     """
    #     Implementation of download_file_query_item with checks for a multi-threaded download.
    #     """
    #     # ckan.session_reset()
    #     # ckan.identifier = current_thread().name
    #     ckan = self.thread_ckan[current_thread().name]
    #     total = self.get_file_query_len()
    #     end_index = positive_end_index(end_index, total)
    #     if self.stop_event.is_set():
    #         return
    #     if external_stop_event is not None and external_stop_event.is_set():
    #         print(f"{ckan.identifier} Interrupted")
    #         return
    #     try:
    #         # self._unit_download_apply(ckan=ckan, file_query_item=file_query_item,
    #         #                           index=index, start_index=start_index, end_index=end_index, total=total)
    #     except Exception as e:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print(f"Stopping all threads because an exception occurred in thread: {e}")
    #         raise e from e

    # def download_request_full_multi_threaded(self, ckan: CkanApi, out_dir: str,
    #                                          threads: int = None, external_stop_event=None,
    #                                          start_index:int=0, end_index:int=-1) -> None:
    #     """
    #     Multi-threaded implementation of download_request_full using ThreadPoolExecutor.
    #     """
    #     self.init_download_file_query_list(ckan=ckan, out_dir=out_dir, cancel_if_present=True)
    #     self._prepare_for_multithreading(ckan)
    #     try:
    #         with ThreadPoolExecutor(max_workers=threads, initializer=self._init_thread, initargs=(ckan,)) as executor:
    #             if ckan.verbose_extra:
    #                 print(f"Launching multi-threaded download of multi-file resource {self.name}")
    #             futures = [executor.submit(self.download_file_query_item_graceful, ckan=ckan, out_dir=out_dir, file_query_item=file_query_item,
    #                                        index=index, external_stop_event=external_stop_event, start_index=start_index, end_index=end_index)
    #                        for index, file_query_item in enumerate(self.get_file_query_generator())]
    #             for future in futures:
    #                 future.result()  # This will propagate the exception
    #         total = self.get_file_query_len()
    #         self._call_progress_callback(total, total, context=f"multi-thread download")
    #     except Exception as e:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print(f"Stopping all threads because an exception occurred: {e}")
    #         raise e from e
    #     finally:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print("End of multi-threaded download...")

    def download_sample_df(self, ckan: CkanApi, search_all:bool=False, **kwargs) -> pd.DataFrame:
        # alias with search_all=False by default
        return super().download_sample_df(ckan=ckan, search_all=search_all, **kwargs)

    def download_sample(self, ckan:CkanApi, full_download:bool=False, **kwargs) -> bytes:
        # alias with full_download=False by default
        return super().download_sample(ckan=ckan, full_download=full_download, **kwargs)





#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements functions to initiate a DataStore.
"""
from abc import ABC, abstractmethod
from typing import Dict, List, Tuple, Union, Set, Any
import os
import io
from warnings import warn
from collections import OrderedDict
import copy

import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.builder.builder_field import BuilderField
from ckanapi_harvesters.harvesters.file_formats.file_format_abc import FileFormatABC
from ckanapi_harvesters.harvesters.file_formats.file_format_init import init_file_format_datastore
from ckanapi_harvesters.builder.mapper_datastore import DataSchemeConversion
from ckanapi_harvesters.builder.builder_resource import BuilderResourceABC
from ckanapi_harvesters.auxiliary.ckan_errors import DuplicateNameError
from ckanapi_harvesters.auxiliary.path import resolve_rel_path
from ckanapi_harvesters.builder.builder_errors import RequiredDataFrameFieldsError, ResourceFileNotExistMessage, IncompletePatchError
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo, CkanDataStoreInfo
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, find_duplicates, datastore_id_col
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.auxiliary.ckan_model import UpsertChoice
from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_abc import CkanDataCleanerABC

# number of rows to upload to initiate DataStore with datapusher, before explicitly specifying field data types and indexes
num_rows_patch_first_upload_partial: Union[int,None] = 50  # set to None to upload directly the whole DataFrame before the DataStore creation


default_alias_keyword:Union[str,None] = "default"  # generate default alias if an alias with this value is found in parameters


class BuilderDataStoreABC(BuilderResourceABC, ABC):
    def __init__(self, *, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None):
        super().__init__(name=name, format=format, description=description, resource_id=resource_id, download_url=download_url)
        self.field_builders: Union[Dict[str, BuilderField],None] = None
        self.primary_key: Union[List[str],None] = None
        self.indexes: Union[List[str],None] = None
        self.aliases: Union[List[str],None] = None
        self.aux_upload_fun_name:str = ""
        self.aux_download_fun_name:str = ""
        # Functions input/outputs
        self.data_cleaner_upload: Union[CkanDataCleanerABC,None] = None
        self.reupload_on_update = False  # do not reupload on update for DataStores
        self.reupload_if_needed: bool = True
        self.reupload_needed: Union[bool,None] = None
        self.df_mapper = DataSchemeConversion()
        self.local_file_format: FileFormatABC = init_file_format_datastore(self.format)

    def copy(self, *, dest=None):
        super().copy(dest=dest)
        dest.field_builders = copy.deepcopy(self.field_builders)
        dest.primary_key = copy.deepcopy(self.primary_key)
        dest.indexes = copy.deepcopy(self.indexes)
        dest.aliases = copy.deepcopy(self.aliases)
        dest.aux_upload_fun_name = self.aux_upload_fun_name
        dest.aux_download_fun_name = self.aux_download_fun_name
        dest.reupload_on_update = self.reupload_on_update
        dest.reupload_if_needed = self.reupload_if_needed
        dest.reupload_needed = self.reupload_needed
        dest.df_mapper = self.df_mapper.copy()
        dest.local_file_format = self.local_file_format.copy()
        return dest

    def _init_file_format(self):
        self.local_file_format = init_file_format_datastore(self.format)  # default file format is CSV (user can change)

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row)
        primary_keys_string: Union[str,None] = _string_from_element(row["primary key"])
        indexes_string: Union[str,None] = _string_from_element(row["indexes"])
        aliases_string: Union[str,None] = None
        if "upload function" in row.keys():
            self.aux_upload_fun_name: str = _string_from_element(row["upload function"], empty_value="")
        if "download function" in row.keys():
            self.aux_download_fun_name: str = _string_from_element(row["download function"], empty_value="")
        if "aliases" in row.keys():
            aliases_string = _string_from_element(row["aliases"])
        if primary_keys_string is not None:
            if primary_keys_string.lower() == "none":
                self.primary_key = []
            else:
                self.primary_key = [field.strip() for field in primary_keys_string.split(ckan_tags_sep)]
        if indexes_string is not None:
            if indexes_string.lower() == "none":
                self.indexes = []
            else:
                self.indexes = [field.strip() for field in indexes_string.split(ckan_tags_sep)]
        if aliases_string is not None:
            self.aliases = aliases_string.split(ckan_tags_sep)
        self._init_file_format()

    @abstractmethod
    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["Primary key"] = ckan_tags_sep.join(self.primary_key) if self.primary_key else ""
        d["Indexes"] = ckan_tags_sep.join(self.indexes) if self.indexes is not None else ""
        d["Upload function"] = self.aux_upload_fun_name
        d["Download function"] = self.aux_download_fun_name
        d["Aliases"] = ckan_tags_sep.join(self.aliases) if self.aliases is not None else ""
        return d

    def init_options_from_ckan(self, ckan:CkanApi) -> None:
        super().init_options_from_ckan(ckan)
        if self.field_builders is not None:
            for field_builder in self.field_builders.values():
                field_builder.internal_attrs.update_from_ckan(ckan)

    def _check_field_duplicates(self):
        if self.field_builders is not None:
            duplicates = find_duplicates([field_builder.name for field_builder in self.field_builders.values()])
            if len(duplicates) > 0:
                raise DuplicateNameError("Field", duplicates)

    def _get_fields_dict(self) -> Dict[str, dict]:
        self._check_field_duplicates()
        if self.field_builders is not None:
            fields_dict = OrderedDict([(field_builder.name, field_builder._to_dict()) for field_builder in self.field_builders.values()])
        else:
            fields_dict = None
        return fields_dict

    def _get_fields_info(self) -> Dict[str, CkanField]:
        self._check_field_duplicates()
        if self.field_builders is not None:
            builder_fields = OrderedDict([(field_builder.name, field_builder._to_ckan_field()) for field_builder in self.field_builders.values()])
        else:
            builder_fields = {}
        return builder_fields

    def _get_fields_df(self) -> pd.DataFrame:
        fields_dict_list = [value for value in self._get_fields_dict().values()]
        fields_df = pd.DataFrame.from_records(fields_dict_list)
        return fields_df

    def _load_fields_df(self, fields_df: pd.DataFrame):
        fields_df.columns = fields_df.columns.map(str.lower)
        fields_df.columns = fields_df.columns.map(str.strip)
        self.field_builders = {}
        for index, row in fields_df.iterrows():
            field_builder = BuilderField()
            field_builder._load_from_df_row(row=row)
            self.field_builders[field_builder.name] = field_builder

    def _to_ckan_resource_info(self, package_id:str, check_id:bool=True) -> CkanResourceInfo:
        resource_info = super()._to_ckan_resource_info(package_id=package_id, check_id=check_id)
        resource_info.datastore_info = CkanDataStoreInfo()
        resource_info.datastore_info.resource_id = resource_info.id
        if self.field_builders is not None:
            resource_info.datastore_info.fields_dict = OrderedDict()
            for name, field_builder in self.field_builders.items():
                resource_info.datastore_info.fields_dict[name] = field_builder._to_ckan_field()
        else:
            resource_info.datastore_info.fields_dict = None
        resource_info.datastore_info.fields_id_list = [name for name, field_builder in self.field_builders.items()] if self.field_builders is not None else []
        if self.indexes is not None:
            resource_info.datastore_info.index_fields = self.indexes.copy()
        aliases = self._get_alias_list(None)
        if aliases is not None:
            resource_info.datastore_info.aliases = aliases.copy()
        return resource_info

    @abstractmethod
    def load_sample_df(self, resources_base_dir:str, *, upload_alter:bool=True) -> pd.DataFrame:
        """
        Function returning the data from the indicated resources as a pandas DataFrame.
        This is the DataFrame equivalent for load_sample_data.

        :param resources_base_dir: base directory to find the resources on the local machine
        :return:
        """
        raise NotImplementedError()

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return False

    def get_sample_file_path(self, resources_base_dir: str) -> None:
        return None

    def load_sample_data(self, resources_base_dir:str) -> bytes:
        df = self.load_sample_df(resources_base_dir=resources_base_dir)
        return self.local_file_format.write_in_memory(df, fields=self._get_fields_info())

    def upsert_request_df(self, ckan: CkanApi, df_upload:pd.DataFrame,
                          method:UpsertChoice=UpsertChoice.Upsert,
                          apply_last_condition:bool=None, always_last_condition:bool=None) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Call to ckan datastore_upset.
        Before sending the DataFrame, a call to df_upload_alter is made.
        This method is overloaded in BuilderDataStoreMultiABC and BuilderDataStoreFolder

        :param ckan:
        :param df_upload:
        :param method:
        :return:
        """
        resource_id = self.get_or_query_resource_id(ckan, error_not_found=True)
        df_upload_transformed = self.df_mapper.df_upload_alter(df_upload, fields=self._get_fields_info())
        ret_df = ckan.datastore_upsert(df_upload_transformed, resource_id, method=method,
                                       apply_last_condition=apply_last_condition,
                                       always_last_condition=always_last_condition, data_cleaner=self.data_cleaner_upload)
        return df_upload_transformed, ret_df

    def upsert_request_final(self, ckan: CkanApi, *, force:bool=False) -> None:
        """
        Final steps after the last upsert query.
        These steps are automatically done for a DataStore defined by one file.

        :param ckan:
        :param force: perform request anyways
        :return:
        """
        if force:
            resource_id = self.get_or_query_resource_id(ckan, error_not_found=True)
            ckan.datastore_upsert_last_line(resource_id=resource_id)

    def _get_alias_list(self, ckan:Union[CkanApi,None]):
        aliases = self.aliases
        if default_alias_keyword is not None:
            if ckan is not None:
                default_alias_name = ckan.datastore_default_alias(self.name, self.package_name, error_not_found=False)
            else:
                default_alias_name = CkanApi.datastore_default_alias_of_names(self.name, self.package_name)
            if aliases is not None:
                for i, alias in enumerate(aliases):
                    if alias.lower().strip() == default_alias_keyword:
                        aliases[i] = default_alias_name
        return aliases

    def _check_necessary_fields(self, current_fields: Set[str] = None, empty_datastore:bool=False, raise_error: bool = True) -> Set[str]:
        """
        Auxiliary function to list the fields which are required:
        - for df_mapper to determine the file names, associated requests, and recognize the last inserted row of a document.
        - to initialize the DataStore with the columns for the primary key and indexes

        The required fields are compared to current_fields, if provided.
        """
        if empty_datastore:
            return set()
        required_fields = self.df_mapper.get_necessary_fields()
        if self.primary_key is not None:
            required_fields = required_fields.union(set(self.primary_key))
        if self.indexes is not None:
            required_fields = required_fields.union(set(self.indexes))
        if current_fields is not None:
            missing_fields = required_fields - current_fields
            if len(missing_fields) > 0:
                msg = RequiredDataFrameFieldsError(missing_fields)
                if raise_error:
                    raise msg
                else:
                    warn(str(msg))
        return required_fields

    def _check_undocumented_fields(self, current_fields: Set[str]) -> None:
        if self.field_builders is not None:
            # list fields which are not documented
            fields_doc = set(self.field_builders.keys())
            missing_doc = current_fields - fields_doc
            extra_doc = fields_doc - current_fields
            if len(extra_doc) > 0:
                msg = f"{len(extra_doc)} extra fields were documented but absent of sample data for table {self.name}: {', '.join(extra_doc)}"
                warn(msg)
            if len(missing_doc) > 0:
                msg = f"{len(missing_doc)} fields are left documented for table {self.name}: {', '.join(missing_doc)}"
                warn(msg)
        else:
            msg = f"No field documentation was provided for table {self.name}. {len(current_fields)} fields are left documented: {', '.join(current_fields)}"
            warn(msg)

    def _get_fields_update(self, ckan: CkanApi, current_fields:Union[Set[str],None], data_cleaner_fields:Union[List[dict],None],
                           reupload:bool) -> Dict[str, dict]:
        if self.field_builders is not None:
            if current_fields is not None:
                builder_fields = [field_builder._to_ckan_field() for field_builder in self.field_builders.values() if field_builder.name in current_fields]
            else:
                # use case: get all known fields (before data_cleaner)
                builder_fields = [field_builder._to_ckan_field() for field_builder in self.field_builders.values()]
        else:
            builder_fields = None
        resource_id = self.get_or_query_resource_id(ckan, error_not_found=False)
        if resource_id is not None and not reupload:
            update_needed, fields_update = ckan.datastore_field_patch_dict(fields_merge=data_cleaner_fields, fields_update=builder_fields,
                                                                           return_list=False,
                                                                           resource_id=resource_id, error_not_found=False)
        else:
            fields_update = CkanApi.datastore_field_dict(fields_merge=data_cleaner_fields, fields_update=builder_fields, return_list=False)
        return fields_update

    def _collect_indexes_from_fields(self) -> Set[str]:
        if self.field_builders is not None:
            return {field_builder.name for field_builder in self.field_builders.values() if field_builder.is_index}
        else:
            return set()

    def _get_primary_key_indexes(self, data_cleaner_index: Set[str], current_fields:Set[str], error_missing:bool, empty_datastore:bool=False) -> Tuple[Union[List[str],None], Union[List[str],None]]:
        # update primary keys and indexes: only if present
        if empty_datastore:
            return None, None
        primary_key = None
        if current_fields is None:
            primary_key = self.primary_key
        elif self.primary_key is not None:
            extra_primary_key = set(self.primary_key) - current_fields
            if len(extra_primary_key) == 0:
                primary_key = self.primary_key
            elif error_missing:
                raise RequiredDataFrameFieldsError(extra_primary_key)
        indexes = None
        if self.indexes is not None:
            indexes_full_set = set(self.indexes).union(self._collect_indexes_from_fields()).union(data_cleaner_index)
        else:
            indexes_full_set = self._collect_indexes_from_fields().union(data_cleaner_index)
        if primary_key is not None:
            indexes_full_set = indexes_full_set - set(primary_key)
        if len(indexes_full_set) == 0:
            indexes_full = None
        else:
            indexes_full = list(indexes_full_set)
        if current_fields is None:
            indexes = indexes_full
        elif indexes_full is not None:
            extra_indexes = set(indexes_full) - current_fields
            if len(extra_indexes) == 0:
                indexes = indexes_full
            elif error_missing:
                raise RequiredDataFrameFieldsError(extra_indexes)
        return primary_key, indexes

    def _compare_fields_to_datastore_info(self, resource_info:CkanResourceInfo, current_fields: Set[str], ckan:CkanApi) -> None:
        # compare fields with DataStore info (if present, for information)
        if resource_info.datastore_info is not None:
            fields_info = set(resource_info.datastore_info.fields_id_list)
            missing_info = current_fields - fields_info
            extra_info = fields_info - current_fields
            if len(extra_info) > 0:
                msg = f"{len(extra_info)} extra fields are in the database but absent of sample data for table {self.name}: {', '.join(extra_info)}"
                warn(msg)
            if len(missing_info) > 0 and ckan.params.verbose_request:
                msg = f"{len(missing_info)} fields are not in DataStore info because they are being added for table {self.name}: {', '.join(missing_info)}"
                print(msg)

    def _apply_data_cleaner_before_patch(self, ckan:CkanApi, df_upload: pd.DataFrame, reupload:bool) -> Tuple[pd.DataFrame, List[dict], Set[str]]:
        if df_upload is not None and self.data_cleaner_upload is not None:
            fields_for_cleaner = self._get_fields_update(ckan, current_fields=None, data_cleaner_fields=None, reupload=reupload)
            df_upload = self.data_cleaner_upload.clean_records(df_upload, known_fields=fields_for_cleaner, inplace=True)
            data_cleaner_fields = self.data_cleaner_upload.merge_field_changes()
            data_cleaner_index = self.data_cleaner_upload.field_suggested_index
        else:
            data_cleaner_fields = None
            data_cleaner_index = set()
        return df_upload, data_cleaner_fields, data_cleaner_index

    def patch_request(self, ckan: CkanApi, package_id: str, *,
                      df_upload: pd.DataFrame=None, reupload: bool = None, resources_base_dir:str=None) -> CkanResourceInfo:
        if reupload is None: reupload = self.reupload_on_update
        if df_upload is None:
            df_upload = self.load_sample_df(resources_base_dir=resources_base_dir, upload_alter=True)
        else:
            pass  # do not alter df_upload because it should already be in the database format
        df_upload, data_cleaner_fields, data_cleaner_index = self._apply_data_cleaner_before_patch(ckan, df_upload, reupload=reupload)
        current_fields = set(df_upload.columns) - {datastore_id_col}  # _id field cannot be documented
        if num_rows_patch_first_upload_partial is not None and len(df_upload) > num_rows_patch_first_upload_partial:
            df_upload_partial = df_upload.iloc[:num_rows_patch_first_upload_partial]
            df_upload_upsert = df_upload.iloc[num_rows_patch_first_upload_partial:]
        else:
            df_upload_partial, df_upload_upsert = df_upload, None
        empty_datastore = df_upload is None or len(df_upload) == 0
        self._check_necessary_fields(current_fields, empty_datastore=empty_datastore, raise_error=True)
        self._check_undocumented_fields(current_fields)
        aliases = self._get_alias_list(ckan)
        primary_key, indexes = self._get_primary_key_indexes(data_cleaner_index, current_fields=current_fields,
                                                             error_missing=True, empty_datastore=empty_datastore)
        fields_update = self._get_fields_update(ckan, current_fields, data_cleaner_fields, reupload=reupload)
        fields = list(fields_update.values()) if len(fields_update) > 0 else None
        resource_info = ckan.resource_create(package_id, name=self.name, format=self.format, description=self.description, state=self.state,
                                             create_default_view=self.create_default_view,
                                             cancel_if_exists=True, update_if_exists=True, reupload=reupload,
                                             datastore_create=True, records=df_upload_partial, fields=fields,
                                             primary_key=primary_key, indexes=indexes, aliases=aliases)
        resource_id = resource_info.id
        self.known_id = resource_id
        reupload = reupload or resource_info.newly_created
        self._compare_fields_to_datastore_info(resource_info, current_fields, ckan)
        if df_upload_upsert is not None and reupload:
            if reupload:
                ckan.datastore_upsert(df_upload_upsert, resource_id, method=UpsertChoice.Insert,
                                      always_last_condition=None, data_cleaner=self.data_cleaner_upload, )
            else:
                # case where a reupload was needed but is not permitted by self.reupload_if_needed
                msg = f"Did not upload the remaining part of the resource {self.name}."
                raise IncompletePatchError(msg)
        return resource_info

    def download_sample_df(self, ckan: CkanApi, search_all:bool=True, download_alter:bool=True, **kwargs) -> Union[pd.DataFrame,None]:
        """
        Download the resource and return it as a DataFrame.
        This is the DataFrame equivalent for download_sample.

        :param ckan:
        :param search_all:
        :param download_alter:
        :param kwargs:
        :return:
        """
        resource_id = self.get_or_query_resource_id(ckan=ckan, error_not_found=self.download_error_not_found)
        if resource_id is None and not self.download_error_not_found:
            return None
        df_download = ckan.datastore_dump(resource_id, search_all=search_all, **kwargs)
        if download_alter:
            df_local = self.df_mapper.df_download_alter(df_download, fields=self._get_fields_info())
            return df_local
        else:
            return df_download

    def download_sample(self, ckan:CkanApi, full_download:bool=True, **kwargs) -> bytes:
        df = self.download_sample_df(ckan=ckan, search_all=full_download, **kwargs)
        return self.local_file_format.write_in_memory(df, fields=self._get_fields_info())


class BuilderDataStoreFile(BuilderDataStoreABC):
    def __init__(self, *, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, file_name:str=None):
        super().__init__(name=name, format=format, description=description, resource_id=resource_id, download_url=download_url)
        self.file_name = file_name

    def copy(self, *, dest=None):
        if dest is None:
            dest = BuilderDataStoreFile()
        super().copy(dest=dest)
        dest.file_name = self.file_name
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row)
        self.file_name: str = _string_from_element(row["file/url"])

    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        file_path = self.get_sample_file_path(resources_base_dir=resources_base_dir)
        if os.path.isfile(file_path):
            return None
        else:
            return ResourceFileNotExistMessage(self.name, ErrorLevel.Error, f"Missing file for resource {self.name}: {file_path}")

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return False

    def get_sample_file_path(self, resources_base_dir:str) -> str:
        return resolve_rel_path(resources_base_dir, self.file_name, field=f"File/URL of resource {self.name}")

    def load_sample_df(self, resources_base_dir:str, *, upload_alter:bool=True) -> pd.DataFrame:
        self.sample_data_source = self.get_sample_file_path(resources_base_dir)
        df_local = self.local_file_format.read_file(self.sample_data_source, fields=self._get_fields_info())
        if isinstance(df_local, pd.DataFrame):
            df_local.attrs["source"] = self.sample_data_source
        if upload_alter:
            df_upload = self.df_mapper.df_upload_alter(df_local, self.sample_data_source, fields=self._get_fields_info())
            return df_upload
        else:
            return df_local

    @staticmethod
    def resource_mode_str() -> str:
        return "DataStore from File"

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = self.file_name
        return d

    def download_request(self, ckan: CkanApi, out_dir: str, *, full_download:bool=True,
                         force:bool=False, threads:int=1) -> Union[pd.DataFrame,None]:
        if (not self.enable_download) and (not force):
            msg = f"Did not download resource {self.name} because download was disabled."
            warn(msg)
            return None
        if out_dir is not None:
            self.downloaded_destination = resolve_rel_path(out_dir, self.file_name, field=f"File/URL of resource {self.name}")
            if self.download_skip_existing and os.path.exists(self.downloaded_destination):
                return None
        resource_id = self.get_or_query_resource_id(ckan=ckan, error_not_found=self.download_error_not_found)
        if resource_id is None and not self.download_error_not_found:
            return None
        df_download = ckan.datastore_dump(resource_id, search_all=full_download)
        df = self.df_mapper.df_download_alter(df_download, fields=self._get_fields_info())
        if out_dir is not None:
            os.makedirs(out_dir, exist_ok=True)
            self.local_file_format.write_file(df, self.downloaded_destination, fields=self._get_fields_info())
        return df


class BuilderResourceIgnored(BuilderDataStoreABC):
    """
    Class to maintain a line in the resource builders list but has no action and can hold field metadata.
    """
    def __init__(self, *, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, file_url:str=None):
        super().__init__(name=name, format=format, description=description, resource_id=resource_id, download_url=download_url)
        self.file_url: Union[str, None] = file_url

    def copy(self, *, dest=None):
        if dest is None:
            dest = BuilderResourceIgnored()
        super().copy(dest=dest)
        dest.file_url = self.file_url
        return dest

    @staticmethod
    def resource_mode_str() -> str:
        return "Ignored"

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row)
        self.file_url: str = _string_from_element(row["file/url"])
        self._check_mandatory_attributes()

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = self.file_url
        return d

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return False

    def get_sample_file_path(self, resources_base_dir:str) -> Union[str,None]:
        return None

    def load_sample_data(self, resources_base_dir:str) -> Union[bytes,None]:
        return None

    def load_sample_df(self, resources_base_dir: str, *, upload_alter: bool = True) -> None:
        return None

    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[ContextErrorLevelMessage,None]:
        return None

    def patch_request(self, ckan:CkanApi, package_id:str, *,
                      reupload:bool=None, resources_base_dir:str=None,
                      payload:Union[bytes, io.BufferedIOBase]=None) -> None:
        return None

    def download_request(self, ckan: CkanApi, out_dir: str, *, full_download: bool = True, force: bool = False,
                         threads: int = 1) -> Any:
        return None

    def download_sample(self, ckan: CkanApi, full_download: bool = True, **kwargs) -> bytes:
        return bytes()




#!python3
# -*- coding: utf-8 -*-
"""
Package to enforce CKAN data policies.
"""

POLICY_FILE_FORMAT_VERSION = "0.0.0"  # version of the data format policy file format

from . import data_format_policy_defs
from . import data_format_policy_errors
from . import data_format_policy_abc
from . import data_format_policy_lists
from . import data_format_policy_tag_groups
from . import data_format_policy_custom_fields
from . import data_format_policy

# usage shortcuts
from ckanapi_harvesters.policies.data_format_policy import CkanPackageDataFormatPolicy





#!python3
# -*- coding: utf-8 -*-
"""
Data format policy representation and enforcing
"""
from typing import List, Tuple
from warnings import warn
from collections import OrderedDict

from ckanapi_harvesters.auxiliary.error_level_message import ErrorLevelMessage, ErrorLevel


class DataPolicyError(ErrorLevelMessage):
    def __init__(self, context:str, error_level:ErrorLevel, policy_message: str):
        message = f"In {context} / Data format policy {error_level.name}: {policy_message}"
        super().__init__(error_level, message)
        self.context: str = context
        self.specific_message: str = policy_message

    def to_dict(self) -> dict:
        return OrderedDict([
            ("level", str(self.error_level)),
            ("context", self.context),
            ("message", self.specific_message),
        ])


class UnsupportedPolicyVersionError(Exception):
    def __init__(self, file_version):
        super().__init__(f"Version error: policy file version {file_version} is not supported")

class UrlPolicyLockedError(Exception):
    def __init__(self, url):
        super().__init__(f"Url is not allowed a policy definition - feature locked (url: {url})")

def _policy_msg(msg:DataPolicyError, *, error_level:ErrorLevel, buffer:List[DataPolicyError], verbose:bool) -> None:
    if buffer is not None:
        buffer.append(msg)
    elif error_level == ErrorLevel.Information and verbose:
        print(str(msg))
    elif error_level == ErrorLevel.Warning:
        msg = str(msg)
        warn(msg)
    elif error_level == ErrorLevel.Error:
        raise msg


class ErrorCount:
    def __init__(self, messages_list:List[DataPolicyError]):
        self.messages_list:List[DataPolicyError] = messages_list
        self.information:int = 0
        self.warning:int = 0
        self.error:int = 0
        self.total:int = len(messages_list)
        for message in messages_list:
            if message.error_level == ErrorLevel.Information:
                self.information += 1
            elif message.error_level == ErrorLevel.Warning:
                self.warning += 1
            elif message.error_level == ErrorLevel.Error:
                self.error += 1

    def error_count_message(self) -> str:
        if self.total == 0:
            return "All tests passed"
        else:
            return f"{self.error} errors, {self.warning} warnings, {self.information} messages"

    def __str__(self) -> str:
        return "ErrorCount: " + self.error_count_message()

    def __add__(self, other):
        return ErrorCount(self.messages_list + other.messages_list)

    def to_tuple(self) -> Tuple[int, int, int]:
        return (self.error, self.warning, self.information)

    def to_dict(self) -> dict[str,int]:
        return OrderedDict([("errors", self.error), ("warnings", self.warning), ("information", self.information)])



#!python3
# -*- coding: utf-8 -*-
"""
Methods to load an API key
"""

import os.path
from warnings import warn
from typing import Dict, Union, Iterable
import getpass
import argparse

from ckanapi_harvesters.auxiliary.ckan_errors import ApiKeyFileError
from ckanapi_harvesters.auxiliary.path import sanitize_path, path_rel_to_dir
from ckanapi_harvesters.auxiliary.ckan_defs import environ_keyword



class ApiKey:
    """
    API key storage class.
    """
    CKAN_API_KEY_HEADER_NAME = {"Authorization", "X-CKAN-API-Key"}  # match apikey_header_name of your CKAN instance
    CKAN_API_KEY_ENVIRON = "CKAN_API_KEY"  # not recommended to store sensitive information in environment variables
    API_KEY_FILE_ENVIRON = "CKAN_API_KEY_FILE"

    def __init__(self, *, apikey:str=None, apikey_file:str=None,
                 api_key_header_name:Union[str, Iterable[str]]=None):
        """
        CKAN Database API key storage class.

        :param apikey: way to provide the API key directly (optional)
        :param apikey_file: path to a file containing a valid API key in the first line of text (optional)
        """
        if api_key_header_name is None:
            api_key_header_name = "Authorization"
        self.apikey_file: str = apikey_file  # path to a file containing a valid API key in the first line of text (optional)
        self._apikey: str = apikey  # API key used for restricted package access
        self.api_key_header_name = api_key_header_name

    def __del__(self):
        self.clear()

    def __copy__(self):
        return self.copy()

    def copy(self, *, dest=None):
        if dest is None:
            dest = ApiKey()
        dest.apikey_file = self.apikey_file
        dest._apikey = self._apikey
        return dest

    def __str__(self):
        if self._apikey is None:
            return "None"
        elif self._apikey == "":
            return "<empty string>"
        else:
            return "*****"

    @property
    def value(self) -> Union[str,None]:
        return self._apikey
    @value.setter
    def value(self, value:Union[str,None]):
        self._apikey = value

    def is_empty(self):
        return self._apikey is None

    def clear(self) -> None:
        self._apikey = None

    def load_from_environ(self, *, error_not_found:bool=False) -> bool:
        """
        Load CKAN API key from environment variables, by order of priority:

        By default, no environment variables are used.
        """
        return False

    def load_apikey(self, apikey_file:str=None, *, base_dir:str=None, error_not_found:bool=True) -> bool:
        """
        Load the API key from file.
        The file should contain a valid API key in the first line of text.

        :param apikey_file: path to the API key file. The following keywords are accepted:
            - "environ": the API key will be looked up in the environment variable with load_from_environ
        :param base_dir: base directory to find the API key file, if a relative path is provided
        :param error_not_found: option to raise an exception if the API key file is not found
        :return:
        """
        if apikey_file is None:
            apikey_file = self.apikey_file
        apikey_file = path_rel_to_dir(apikey_file, base_dir=base_dir, keyword_exceptions={environ_keyword})
        if apikey_file is None:
            raise ApiKeyFileError('apikey_file is required')
        api_keyword = apikey_file.strip().lower()
        if api_keyword == environ_keyword:
            return self.load_from_environ(error_not_found=error_not_found)
        if not(os.path.isfile(apikey_file)) and not error_not_found:
            msg = f"API key file does not exist: {apikey_file}"
            warn(msg)
            return False
        with open(apikey_file, 'r') as f:
            apikey = f.readline().strip()
            f.close()
        self.value = apikey
        self.apikey_file = apikey_file
        return True

    def get_auth_header(self) -> Dict[str, str]:
        """
        Returns the correct header with the API key for the requests needing it.
        If no API key was loaded, returns an empty dictionary.
        """
        if self.value is not None:
            apikey_encoded = self.value
            if isinstance(self.api_key_header_name, str):
                return {self.api_key_header_name: apikey_encoded}
            else:
                return {key: apikey_encoded for key in self.api_key_header_name}
        else:
            return {}

    def input(self):
        """
        Prompt the user to input the API key in the console window.

        :return:
        """
        api_key = getpass.getpass("Please enter the API key: ")
        self._apikey = api_key

    @staticmethod
    def _setup_cli_parser(parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        if parser is None:
            parser = argparse.ArgumentParser(description="API key initialization")
        parser.add_argument("--apikey", type=str,
                            help="API key")
        parser.add_argument("--apikey-file", type=str,
                            help="Path to a file containing the API key (first line)")
        return parser

    def _cli_args_apply(self, args: argparse.Namespace, *, base_dir: str = None, error_not_found: bool = True) -> None:
        if args.apikey is not None:
            self.value = args.apikey
        if args.apikey_file is not None:
            self.load_apikey(args.apikey_file, base_dir=base_dir, error_not_found=error_not_found)


class CkanApiKey(ApiKey):
    """
    CKAN Database API key storage class.
    """

    def __init__(self, *, apikey:str=None, apikey_file:str=None):
        """
        CKAN Database API key storage class.

        :param apikey: way to provide the API key directly (optional)
        :param apikey_file: path to a file containing a valid API key in the first line of text (optional)
        """
        super().__init__(apikey=apikey, apikey_file=apikey_file, api_key_header_name=self.CKAN_API_KEY_HEADER_NAME)

    def copy(self, *, dest=None) -> "CkanApiKey":
        if dest is None:
            dest = CkanApiKey()
        super().copy(dest=dest)
        return dest

    def load_from_environ(self, *, error_not_found:bool=False) -> bool:
        """
        Load CKAN API key from environment variables, by order of priority:

        - `CKAN_API_KEY`: for the raw API key (it is not recommended to store API key in an environment variable)
        - `CKAN_API_KEY_FILE`: path to a file containing a valid API key in the first line of text

        :param error_not_found: raise an error if the API key file was not found
        :return:
        """
        apikey = os.environ.get(self.CKAN_API_KEY_ENVIRON)            # "CKAN_API_KEY"
        apikey_file = sanitize_path(os.environ.get(self.API_KEY_FILE_ENVIRON))  # "CKAN_API_KEY_FILE"
        if apikey is not None:
            msg = f"It is not recommended to store sensitive information in environment variables such as the API key ({self.CKAN_API_KEY_ENVIRON})"
            warn(msg)
            self.value = apikey
            return True
        elif apikey_file is not None:
            assert not apikey_file.strip().lower() == environ_keyword  # this value would create an infinite loop
            return self.load_apikey(apikey_file, error_not_found=error_not_found)
        else:
            msg = f"No API key was found in the environment variable {self.CKAN_API_KEY_ENVIRON}"
            warn(msg)
            return False

    def input(self):
        """
        Prompt the user to input the API key in the console window.

        :return:
        """
        api_key = getpass.getpass("Please enter the CKAN API key: ")
        self._apikey = api_key

    @staticmethod
    def _setup_cli_parser(parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        if parser is None:
            parser = argparse.ArgumentParser(description="CKAN API key initialization")
        ApiKey._setup_cli_parser(parser=parser)
        return parser

